# ------------------------------------------------------------------------
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ------------------------------------------------------------------------
# https://raw.githubusercontent.com/deroneriksson/python-wsi-preprocessing/master/deephistopath/wsi/filter.py
import math
import multiprocessing
import numpy as np
import os
import scipy.ndimage.morphology as sc_morph
import skimage.color as sk_color
import skimage.exposure as sk_exposure
import skimage.feature as sk_feature
import skimage.filters as sk_filters
import skimage.future as sk_future
import skimage.morphology as sk_morphology
import skimage.segmentation as sk_segmentation

# from deephistopath.wsi import slide
# from deephistopath.wsi import util
# from deephistopath.wsi.util import Time


def filter_rgb_to_grayscale(np_img, output_type="uint8"):
  """
  Convert an RGB NumPy array to a grayscale NumPy array.

  Shape (h, w, c) to (h, w).

  Args:
    np_img: RGB Image as a NumPy array.
    output_type: Type of array to return (float or uint8)

  Returns:
    Grayscale image as NumPy array with shape (h, w).
  """
  t = Time()
  # Another common RGB ratio possibility: [0.299, 0.587, 0.114]
  grayscale = np.dot(np_img[..., :3], [0.2125, 0.7154, 0.0721])
  if output_type != "float":
    grayscale = grayscale.astype("uint8")
  util.np_info(grayscale, "Gray", t.elapsed())
  return grayscale


def filter_complement(np_img, output_type="uint8"):
  """
  Obtain the complement of an image as a NumPy array.

  Args:
    np_img: Image as a NumPy array.
    type: Type of array to return (float or uint8).

  Returns:
    Complement image as Numpy array.
  """
  t = Time()
  if output_type == "float":
    complement = 1.0 - np_img
  else:
    complement = 255 - np_img
  util.np_info(complement, "Complement", t.elapsed())
  return complement


def filter_hysteresis_threshold(np_img, low=50, high=100, output_type="uint8"):
  """
  Apply two-level (hysteresis) threshold to an image as a NumPy array, returning a binary image.

  Args:
    np_img: Image as a NumPy array.
    low: Low threshold.
    high: High threshold.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where True, 1.0, and 255 represent a pixel above hysteresis threshold.
  """
  t = Time()
  hyst = sk_filters.apply_hysteresis_threshold(np_img, low, high)
  if output_type == "bool":
    pass
  elif output_type == "float":
    hyst = hyst.astype(float)
  else:
    hyst = (255 * hyst).astype("uint8")
  util.np_info(hyst, "Hysteresis Threshold", t.elapsed())
  return hyst


def filter_otsu_threshold(np_img, output_type="uint8"):
  """
  Compute Otsu threshold on image as a NumPy array and return binary image based on pixels above threshold.

  Args:
    np_img: Image as a NumPy array.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where True, 1.0, and 255 represent a pixel above Otsu threshold.
  """
  t = Time()
  otsu_thresh_value = sk_filters.threshold_otsu(np_img)
  otsu = (np_img > otsu_thresh_value)
  if output_type == "bool":
    pass
  elif output_type == "float":
    otsu = otsu.astype(float)
  else:
    otsu = otsu.astype("uint8") * 255
  util.np_info(otsu, "Otsu Threshold", t.elapsed())
  return otsu


def filter_local_otsu_threshold(np_img, disk_size=3, output_type="uint8"):
  """
  Compute local Otsu threshold for each pixel and return binary image based on pixels being less than the
  local Otsu threshold.

  Args:
    np_img: Image as a NumPy array.
    disk_size: Radius of the disk structuring element used to compute the Otsu threshold for each pixel.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where local Otsu threshold values have been applied to original image.
  """
  t = Time()
  local_otsu = sk_filters.rank.otsu(np_img, sk_morphology.disk(disk_size))
  if output_type == "bool":
    pass
  elif output_type == "float":
    local_otsu = local_otsu.astype(float)
  else:
    local_otsu = local_otsu.astype("uint8") * 255
  util.np_info(local_otsu, "Otsu Local Threshold", t.elapsed())
  return local_otsu


def filter_entropy(np_img, neighborhood=9, threshold=5, output_type="uint8"):
  """
  Filter image based on entropy (complexity).

  Args:
    np_img: Image as a NumPy array.
    neighborhood: Neighborhood size (defines height and width of 2D array of 1's).
    threshold: Threshold value.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where True, 1.0, and 255 represent a measure of complexity.
  """
  t = Time()
  entr = sk_filters.rank.entropy(np_img, np.ones((neighborhood, neighborhood))) > threshold
  if output_type == "bool":
    pass
  elif output_type == "float":
    entr = entr.astype(float)
  else:
    entr = entr.astype("uint8") * 255
  util.np_info(entr, "Entropy", t.elapsed())
  return entr


def filter_canny(np_img, sigma=1, low_threshold=0, high_threshold=25, output_type="uint8"):
  """
  Filter image based on Canny algorithm edges.

  Args:
    np_img: Image as a NumPy array.
    sigma: Width (std dev) of Gaussian.
    low_threshold: Low hysteresis threshold value.
    high_threshold: High hysteresis threshold value.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) representing Canny edge map (binary image).
  """
  t = Time()
  can = sk_feature.canny(np_img, sigma=sigma, low_threshold=low_threshold, high_threshold=high_threshold)
  if output_type == "bool":
    pass
  elif output_type == "float":
    can = can.astype(float)
  else:
    can = can.astype("uint8") * 255
  util.np_info(can, "Canny Edges", t.elapsed())
  return can


def mask_percent(np_img):
  """
  Determine the percentage of a NumPy array that is masked (how many of the values are 0 values).

  Args:
    np_img: Image as a NumPy array.

  Returns:
    The percentage of the NumPy array that is masked.
  """
  if (len(np_img.shape) == 3) and (np_img.shape[2] == 3):
    np_sum = np_img[:, :, 0] + np_img[:, :, 1] + np_img[:, :, 2]
    mask_percentage = 100 - np.count_nonzero(np_sum) / np_sum.size * 100
  else:
    mask_percentage = 100 - np.count_nonzero(np_img) / np_img.size * 100
  return mask_percentage


def tissue_percent(np_img):
  """
  Determine the percentage of a NumPy array that is tissue (not masked).

  Args:
    np_img: Image as a NumPy array.

  Returns:
    The percentage of the NumPy array that is tissue.
  """
  return 100 - mask_percent(np_img)


def filter_remove_small_objects(np_img, min_size=3000, avoid_overmask=True, overmask_thresh=95, output_type="uint8"):
  """
  Filter image to remove small objects (connected components) less than a particular minimum size. If avoid_overmask
  is True, this function can recursively call itself with progressively smaller minimum size objects to remove to
  reduce the amount of masking that this filter performs.

  Args:
    np_img: Image as a NumPy array of type bool.
    min_size: Minimum size of small object to remove.
    avoid_overmask: If True, avoid masking above the overmask_thresh percentage.
    overmask_thresh: If avoid_overmask is True, avoid masking above this threshold percentage value.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8).
  """
  t = Time()

  rem_sm = np_img.astype(bool)  # make sure mask is boolean
  rem_sm = sk_morphology.remove_small_objects(rem_sm, min_size=min_size)
  mask_percentage = mask_percent(rem_sm)
  if (mask_percentage >= overmask_thresh) and (min_size >= 1) and (avoid_overmask is True):
    new_min_size = min_size / 2
    print("Mask percentage %3.2f%% >= overmask threshold %3.2f%% for Remove Small Objs size %d, so try %d" % (
      mask_percentage, overmask_thresh, min_size, new_min_size))
    rem_sm = filter_remove_small_objects(np_img, new_min_size, avoid_overmask, overmask_thresh, output_type)
  np_img = rem_sm

  if output_type == "bool":
    pass
  elif output_type == "float":
    np_img = np_img.astype(float)
  else:
    np_img = np_img.astype("uint8") * 255

  util.np_info(np_img, "Remove Small Objs", t.elapsed())
  return np_img


def filter_remove_small_holes(np_img, min_size=3000, output_type="uint8"):
  """
  Filter image to remove small holes less than a particular size.

  Args:
    np_img: Image as a NumPy array of type bool.
    min_size: Remove small holes below this size.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8).
  """
  t = Time()

  rem_sm = sk_morphology.remove_small_holes(np_img, min_size=min_size)

  if output_type == "bool":
    pass
  elif output_type == "float":
    rem_sm = rem_sm.astype(float)
  else:
    rem_sm = rem_sm.astype("uint8") * 255

  util.np_info(rem_sm, "Remove Small Holes", t.elapsed())
  return rem_sm


def filter_contrast_stretch(np_img, low=40, high=60):
  """
  Filter image (gray or RGB) using contrast stretching to increase contrast in image based on the intensities in
  a specified range.

  Args:
    np_img: Image as a NumPy array (gray or RGB).
    low: Range low value (0 to 255).
    high: Range high value (0 to 255).

  Returns:
    Image as NumPy array with contrast enhanced.
  """
  t = Time()
  low_p, high_p = np.percentile(np_img, (low * 100 / 255, high * 100 / 255))
  contrast_stretch = sk_exposure.rescale_intensity(np_img, in_range=(low_p, high_p))
  util.np_info(contrast_stretch, "Contrast Stretch", t.elapsed())
  return contrast_stretch


def filter_histogram_equalization(np_img, nbins=256, output_type="uint8"):
  """
  Filter image (gray or RGB) using histogram equalization to increase contrast in image.

  Args:
    np_img: Image as a NumPy array (gray or RGB).
    nbins: Number of histogram bins.
    output_type: Type of array to return (float or uint8).

  Returns:
     NumPy array (float or uint8) with contrast enhanced by histogram equalization.
  """
  t = Time()
  # if uint8 type and nbins is specified, convert to float so that nbins can be a value besides 256
  if np_img.dtype == "uint8" and nbins != 256:
    np_img = np_img / 255
  hist_equ = sk_exposure.equalize_hist(np_img, nbins=nbins)
  if output_type == "float":
    pass
  else:
    hist_equ = (hist_equ * 255).astype("uint8")
  util.np_info(hist_equ, "Hist Equalization", t.elapsed())
  return hist_equ


def filter_adaptive_equalization(np_img, nbins=256, clip_limit=0.01, output_type="uint8"):
  """
  Filter image (gray or RGB) using adaptive equalization to increase contrast in image, where contrast in local regions
  is enhanced.

  Args:
    np_img: Image as a NumPy array (gray or RGB).
    nbins: Number of histogram bins.
    clip_limit: Clipping limit where higher value increases contrast.
    output_type: Type of array to return (float or uint8).

  Returns:
     NumPy array (float or uint8) with contrast enhanced by adaptive equalization.
  """
  t = Time()
  adapt_equ = sk_exposure.equalize_adapthist(np_img, nbins=nbins, clip_limit=clip_limit)
  if output_type == "float":
    pass
  else:
    adapt_equ = (adapt_equ * 255).astype("uint8")
  util.np_info(adapt_equ, "Adapt Equalization", t.elapsed())
  return adapt_equ


def filter_local_equalization(np_img, disk_size=50):
  """
  Filter image (gray) using local equalization, which uses local histograms based on the disk structuring element.

  Args:
    np_img: Image as a NumPy array.
    disk_size: Radius of the disk structuring element used for the local histograms

  Returns:
    NumPy array with contrast enhanced using local equalization.
  """
  t = Time()
  local_equ = sk_filters.rank.equalize(np_img, selem=sk_morphology.disk(disk_size))
  util.np_info(local_equ, "Local Equalization", t.elapsed())
  return local_equ


def filter_rgb_to_hed(np_img, output_type="uint8"):
  """
  Filter RGB channels to HED (Hematoxylin - Eosin - Diaminobenzidine) channels.

  Args:
    np_img: RGB image as a NumPy array.
    output_type: Type of array to return (float or uint8).

  Returns:
    NumPy array (float or uint8) with HED channels.
  """
  t = Time()
  hed = sk_color.rgb2hed(np_img)
  if output_type == "float":
    hed = sk_exposure.rescale_intensity(hed, out_range=(0.0, 1.0))
  else:
    hed = (sk_exposure.rescale_intensity(hed, out_range=(0, 255))).astype("uint8")

  util.np_info(hed, "RGB to HED", t.elapsed())
  return hed


def filter_rgb_to_hsv(np_img, display_np_info=True):
  """
  Filter RGB channels to HSV (Hue, Saturation, Value).

  Args:
    np_img: RGB image as a NumPy array.
    display_np_info: If True, display NumPy array info and filter time.

  Returns:
    Image as NumPy array in HSV representation.
  """

  if display_np_info:
    t = Time()
  hsv = sk_color.rgb2hsv(np_img)
  if display_np_info:
    util.np_info(hsv, "RGB to HSV", t.elapsed())
  return hsv


def filter_hsv_to_h(hsv, output_type="int", display_np_info=True):
  """
  Obtain hue values from HSV NumPy array as a 1-dimensional array. If output as an int array, the original float
  values are multiplied by 360 for their degree equivalents for simplicity. For more information, see
  https://en.wikipedia.org/wiki/HSL_and_HSV

  Args:
    hsv: HSV image as a NumPy array.
    output_type: Type of array to return (float or int).
    display_np_info: If True, display NumPy array info and filter time.

  Returns:
    Hue values (float or int) as a 1-dimensional NumPy array.
  """
  if display_np_info:
    t = Time()
  h = hsv[:, :, 0]
  h = h.flatten()
  if output_type == "int":
    h *= 360
    h = h.astype("int")
  if display_np_info:
    util.np_info(hsv, "HSV to H", t.elapsed())
  return h


def filter_hsv_to_s(hsv):
  """
  Experimental HSV to S (saturation).

  Args:
    hsv:  HSV image as a NumPy array.

  Returns:
    Saturation values as a 1-dimensional NumPy array.
  """
  s = hsv[:, :, 1]
  s = s.flatten()
  return s


def filter_hsv_to_v(hsv):
  """
  Experimental HSV to V (value).

  Args:
    hsv:  HSV image as a NumPy array.

  Returns:
    Value values as a 1-dimensional NumPy array.
  """
  v = hsv[:, :, 2]
  v = v.flatten()
  return v


def filter_hed_to_hematoxylin(np_img, output_type="uint8"):
  """
  Obtain Hematoxylin channel from HED NumPy array and rescale it (for example, to 0 to 255 for uint8) for increased
  contrast.

  Args:
    np_img: HED image as a NumPy array.
    output_type: Type of array to return (float or uint8).

  Returns:
    NumPy array for Hematoxylin channel.
  """
  t = Time()
  hema = np_img[:, :, 0]
  if output_type == "float":
    hema = sk_exposure.rescale_intensity(hema, out_range=(0.0, 1.0))
  else:
    hema = (sk_exposure.rescale_intensity(hema, out_range=(0, 255))).astype("uint8")
  util.np_info(hema, "HED to Hematoxylin", t.elapsed())
  return hema


def filter_hed_to_eosin(np_img, output_type="uint8"):
  """
  Obtain Eosin channel from HED NumPy array and rescale it (for example, to 0 to 255 for uint8) for increased
  contrast.

  Args:
    np_img: HED image as a NumPy array.
    output_type: Type of array to return (float or uint8).

  Returns:
    NumPy array for Eosin channel.
  """
  t = Time()
  eosin = np_img[:, :, 1]
  if output_type == "float":
    eosin = sk_exposure.rescale_intensity(eosin, out_range=(0.0, 1.0))
  else:
    eosin = (sk_exposure.rescale_intensity(eosin, out_range=(0, 255))).astype("uint8")
  util.np_info(eosin, "HED to Eosin", t.elapsed())
  return eosin


def filter_binary_fill_holes(np_img, output_type="bool"):
  """
  Fill holes in a binary object (bool, float, or uint8).

  Args:
    np_img: Binary image as a NumPy array.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where holes have been filled.
  """
  t = Time()
  if np_img.dtype == "uint8":
    np_img = np_img / 255
  result = sc_morph.binary_fill_holes(np_img)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Binary Fill Holes", t.elapsed())
  return result


def filter_binary_erosion(np_img, disk_size=5, iterations=1, output_type="uint8"):
  """
  Erode a binary object (bool, float, or uint8).

  Args:
    np_img: Binary image as a NumPy array.
    disk_size: Radius of the disk structuring element used for erosion.
    iterations: How many times to repeat the erosion.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where edges have been eroded.
  """
  t = Time()
  if np_img.dtype == "uint8":
    np_img = np_img / 255
  result = sc_morph.binary_erosion(np_img, sk_morphology.disk(disk_size), iterations=iterations)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Binary Erosion", t.elapsed())
  return result


def filter_binary_dilation(np_img, disk_size=5, iterations=1, output_type="uint8"):
  """
  Dilate a binary object (bool, float, or uint8).

  Args:
    np_img: Binary image as a NumPy array.
    disk_size: Radius of the disk structuring element used for dilation.
    iterations: How many times to repeat the dilation.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) where edges have been dilated.
  """
  t = Time()
  if np_img.dtype == "uint8":
    np_img = np_img / 255
  result = sc_morph.binary_dilation(np_img, sk_morphology.disk(disk_size), iterations=iterations)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Binary Dilation", t.elapsed())
  return result


def filter_binary_opening(np_img, disk_size=3, iterations=1, output_type="uint8"):
  """
  Open a binary object (bool, float, or uint8). Opening is an erosion followed by a dilation.
  Opening can be used to remove small objects.

  Args:
    np_img: Binary image as a NumPy array.
    disk_size: Radius of the disk structuring element used for opening.
    iterations: How many times to repeat.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) following binary opening.
  """
  t = Time()
  if np_img.dtype == "uint8":
    np_img = np_img / 255
  result = sc_morph.binary_opening(np_img, sk_morphology.disk(disk_size), iterations=iterations)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Binary Opening", t.elapsed())
  return result


def filter_binary_closing(np_img, disk_size=3, iterations=1, output_type="uint8"):
  """
  Close a binary object (bool, float, or uint8). Closing is a dilation followed by an erosion.
  Closing can be used to remove small holes.

  Args:
    np_img: Binary image as a NumPy array.
    disk_size: Radius of the disk structuring element used for closing.
    iterations: How many times to repeat.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array (bool, float, or uint8) following binary closing.
  """
  t = Time()
  if np_img.dtype == "uint8":
    np_img = np_img / 255
  result = sc_morph.binary_closing(np_img, sk_morphology.disk(disk_size), iterations=iterations)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Binary Closing", t.elapsed())
  return result


def filter_kmeans_segmentation(np_img, compactness=10, n_segments=800):
  """
  Use K-means segmentation (color/space proximity) to segment RGB image where each segment is
  colored based on the average color for that segment.

  Args:
    np_img: Binary image as a NumPy array.
    compactness: Color proximity versus space proximity factor.
    n_segments: The number of segments.

  Returns:
    NumPy array (uint8) representing 3-channel RGB image where each segment has been colored based on the average
    color for that segment.
  """
  t = Time()
  labels = sk_segmentation.slic(np_img, compactness=compactness, n_segments=n_segments)
  result = sk_color.label2rgb(labels, np_img, kind='avg')
  util.np_info(result, "K-Means Segmentation", t.elapsed())
  return result


def filter_rag_threshold(np_img, compactness=10, n_segments=800, threshold=9):
  """
  Use K-means segmentation to segment RGB image, build region adjacency graph based on the segments, combine
  similar regions based on threshold value, and then output these resulting region segments.

  Args:
    np_img: Binary image as a NumPy array.
    compactness: Color proximity versus space proximity factor.
    n_segments: The number of segments.
    threshold: Threshold value for combining regions.

  Returns:
    NumPy array (uint8) representing 3-channel RGB image where each segment has been colored based on the average
    color for that segment (and similar segments have been combined).
  """
  t = Time()
  labels = sk_segmentation.slic(np_img, compactness=compactness, n_segments=n_segments)
  g = sk_future.graph.rag_mean_color(np_img, labels)
  labels2 = sk_future.graph.cut_threshold(labels, g, threshold)
  result = sk_color.label2rgb(labels2, np_img, kind='avg')
  util.np_info(result, "RAG Threshold", t.elapsed())
  return result


def filter_threshold(np_img, threshold, output_type="bool"):
  """
  Return mask where a pixel has a value if it exceeds the threshold value.

  Args:
    np_img: Binary image as a NumPy array.
    threshold: The threshold value to exceed.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing a mask where a pixel has a value (T, 1.0, or 255) if the corresponding input array
    pixel exceeds the threshold value.
  """
  t = Time()
  result = (np_img > threshold)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Threshold", t.elapsed())
  return result


def filter_green_channel(np_img, green_thresh=200, avoid_overmask=True, overmask_thresh=90, output_type="bool"):
  """
  Create a mask to filter out pixels with a green channel value greater than a particular threshold, since hematoxylin
  and eosin are purplish and pinkish, which do not have much green to them.

  Args:
    np_img: RGB image as a NumPy array.
    green_thresh: Green channel threshold value (0 to 255). If value is greater than green_thresh, mask out pixel.
    avoid_overmask: If True, avoid masking above the overmask_thresh percentage.
    overmask_thresh: If avoid_overmask is True, avoid masking above this threshold percentage value.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing a mask where pixels above a particular green channel threshold have been masked out.
  """
#   t = Time()

  g = np_img[:, :, 1]
  gr_ch_mask = (g < green_thresh) & (g > 0)
  mask_percentage = mask_percent(gr_ch_mask)
  if (mask_percentage >= overmask_thresh) and (green_thresh < 255) and (avoid_overmask is True):
    new_green_thresh = math.ceil((255 - green_thresh) / 2 + green_thresh)
    print(
      "Mask percentage %3.2f%% >= overmask threshold %3.2f%% for Remove Green Channel green_thresh=%d, so try %d" % (
        mask_percentage, overmask_thresh, green_thresh, new_green_thresh))
    gr_ch_mask = filter_green_channel(np_img, new_green_thresh, avoid_overmask, overmask_thresh, output_type)
  np_img = gr_ch_mask

  if output_type == "bool":
    pass
  elif output_type == "float":
    np_img = np_img.astype(float)
  else:
    np_img = np_img.astype("uint8") * 255

#   util.np_info(np_img, "Filter Green Channel", t.elapsed())
  return np_img


def filter_red(rgb, red_lower_thresh, green_upper_thresh, blue_upper_thresh, output_type="bool",
               display_np_info=False):
  """
  Create a mask to filter out reddish colors, where the mask is based on a pixel being above a
  red channel threshold value, below a green channel threshold value, and below a blue channel threshold value.

  Args:
    rgb: RGB image as a NumPy array.
    red_lower_thresh: Red channel lower threshold value.
    green_upper_thresh: Green channel upper threshold value.
    blue_upper_thresh: Blue channel upper threshold value.
    output_type: Type of array to return (bool, float, or uint8).
    display_np_info: If True, display NumPy array info and filter time.

  Returns:
    NumPy array representing the mask.
  """
  if display_np_info:
    t = Time()
  r = rgb[:, :, 0] > red_lower_thresh
  g = rgb[:, :, 1] < green_upper_thresh
  b = rgb[:, :, 2] < blue_upper_thresh
  result = ~(r & g & b)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  if display_np_info:
    util.np_info(result, "Filter Red", t.elapsed())
  return result


def filter_red_pen(rgb, output_type="bool", display_np_info=False):
  """
  Create a mask to filter out red pen marks from a slide.

  Args:
    rgb: RGB image as a NumPy array.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing the mask.
  """
#   t = Time()
  result = filter_red(rgb, red_lower_thresh=150, green_upper_thresh=80, blue_upper_thresh=90, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=110, green_upper_thresh=20, blue_upper_thresh=30, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=185, green_upper_thresh=65, blue_upper_thresh=105, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=195, green_upper_thresh=85, blue_upper_thresh=125, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=220, green_upper_thresh=115, blue_upper_thresh=145, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=125, green_upper_thresh=40, blue_upper_thresh=70, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=200, green_upper_thresh=120, blue_upper_thresh=150, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=100, green_upper_thresh=50, blue_upper_thresh=65, display_np_info=display_np_info) & \
           filter_red(rgb, red_lower_thresh=85, green_upper_thresh=25, blue_upper_thresh=45, display_np_info=display_np_info)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
#   util.np_info(result, "Filter Red Pen", t.elapsed())
  return result


def filter_green(rgb, red_upper_thresh, green_lower_thresh, blue_lower_thresh, output_type="bool",
                 display_np_info=False):
  """
  Create a mask to filter out greenish colors, where the mask is based on a pixel being below a
  red channel threshold value, above a green channel threshold value, and above a blue channel threshold value.
  Note that for the green ink, the green and blue channels tend to track together, so we use a blue channel
  lower threshold value rather than a blue channel upper threshold value.

  Args:
    rgb: RGB image as a NumPy array.
    red_upper_thresh: Red channel upper threshold value.
    green_lower_thresh: Green channel lower threshold value.
    blue_lower_thresh: Blue channel lower threshold value.
    output_type: Type of array to return (bool, float, or uint8).
    display_np_info: If True, display NumPy array info and filter time.

  Returns:
    NumPy array representing the mask.
  """
  if display_np_info:
    t = Time()
  r = rgb[:, :, 0] < red_upper_thresh
  g = rgb[:, :, 1] > green_lower_thresh
  b = rgb[:, :, 2] > blue_lower_thresh
  result = ~(r & g & b)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  if display_np_info:
    util.np_info(result, "Filter Green", t.elapsed())
  return result


def filter_green_pen(rgb, output_type="bool"):
  """
  Create a mask to filter out green pen marks from a slide.

  Args:
    rgb: RGB image as a NumPy array.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing the mask.
  """
#   t = Time()
  result = filter_green(rgb, red_upper_thresh=150, green_lower_thresh=160, blue_lower_thresh=140) & \
           filter_green(rgb, red_upper_thresh=70, green_lower_thresh=110, blue_lower_thresh=110) & \
           filter_green(rgb, red_upper_thresh=45, green_lower_thresh=115, blue_lower_thresh=100) & \
           filter_green(rgb, red_upper_thresh=30, green_lower_thresh=75, blue_lower_thresh=60) & \
           filter_green(rgb, red_upper_thresh=195, green_lower_thresh=220, blue_lower_thresh=210) & \
           filter_green(rgb, red_upper_thresh=225, green_lower_thresh=230, blue_lower_thresh=225) & \
           filter_green(rgb, red_upper_thresh=170, green_lower_thresh=210, blue_lower_thresh=200) & \
           filter_green(rgb, red_upper_thresh=20, green_lower_thresh=30, blue_lower_thresh=20) & \
           filter_green(rgb, red_upper_thresh=50, green_lower_thresh=60, blue_lower_thresh=40) & \
           filter_green(rgb, red_upper_thresh=30, green_lower_thresh=50, blue_lower_thresh=35) & \
           filter_green(rgb, red_upper_thresh=65, green_lower_thresh=70, blue_lower_thresh=60) & \
           filter_green(rgb, red_upper_thresh=100, green_lower_thresh=110, blue_lower_thresh=105) & \
           filter_green(rgb, red_upper_thresh=165, green_lower_thresh=180, blue_lower_thresh=180) & \
           filter_green(rgb, red_upper_thresh=140, green_lower_thresh=140, blue_lower_thresh=150) & \
           filter_green(rgb, red_upper_thresh=185, green_lower_thresh=195, blue_lower_thresh=195)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
#   util.np_info(result, "Filter Green Pen", t.elapsed())
  return result


def filter_blue(rgb, red_upper_thresh, green_upper_thresh, blue_lower_thresh, output_type="bool",
                display_np_info=False):
  """
  Create a mask to filter out blueish colors, where the mask is based on a pixel being below a
  red channel threshold value, below a green channel threshold value, and above a blue channel threshold value.

  Args:
    rgb: RGB image as a NumPy array.
    red_upper_thresh: Red channel upper threshold value.
    green_upper_thresh: Green channel upper threshold value.
    blue_lower_thresh: Blue channel lower threshold value.
    output_type: Type of array to return (bool, float, or uint8).
    display_np_info: If True, display NumPy array info and filter time.

  Returns:
    NumPy array representing the mask.
  """
  if display_np_info:
    t = Time()
  r = rgb[:, :, 0] < red_upper_thresh
  g = rgb[:, :, 1] < green_upper_thresh
  b = rgb[:, :, 2] > blue_lower_thresh
  result = ~(r & g & b)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  if display_np_info:
    util.np_info(result, "Filter Blue", t.elapsed())
  return result


def filter_blue_pen(rgb, output_type="bool", display_np_info=False):
  """
  Create a mask to filter out blue pen marks from a slide.

  Args:
    rgb: RGB image as a NumPy array.
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing the mask.
  """
#   t = Time()
  result = filter_blue(rgb, red_upper_thresh=60, green_upper_thresh=120, blue_lower_thresh=190, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=120, green_upper_thresh=170, blue_lower_thresh=200, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=175, green_upper_thresh=210, blue_lower_thresh=230, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=145, green_upper_thresh=180, blue_lower_thresh=210, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=37, green_upper_thresh=95, blue_lower_thresh=160, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=30, green_upper_thresh=65, blue_lower_thresh=130, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=130, green_upper_thresh=155, blue_lower_thresh=180, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=40, green_upper_thresh=35, blue_lower_thresh=85, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=30, green_upper_thresh=20, blue_lower_thresh=65, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=90, green_upper_thresh=90, blue_lower_thresh=140, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=60, green_upper_thresh=60, blue_lower_thresh=120, display_np_info=display_np_info) & \
           filter_blue(rgb, red_upper_thresh=110, green_upper_thresh=110, blue_lower_thresh=175, display_np_info=display_np_info)
  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
#   util.np_info(result, "Filter Blue Pen", t.elapsed())
  return result


def filter_grays(rgb, tolerance=15, output_type="bool"):
  """
  Create a mask to filter out pixels where the red, green, and blue channel values are similar.

  Args:
    np_img: RGB image as a NumPy array.
    tolerance: Tolerance value to determine how similar the values must be in order to be filtered out
    output_type: Type of array to return (bool, float, or uint8).

  Returns:
    NumPy array representing a mask where pixels with similar red, green, and blue values have been masked out.
  """
  t = Time()
  (h, w, c) = rgb.shape

  rgb = rgb.astype(np.int)
  rg_diff = abs(rgb[:, :, 0] - rgb[:, :, 1]) <= tolerance
  rb_diff = abs(rgb[:, :, 0] - rgb[:, :, 2]) <= tolerance
  gb_diff = abs(rgb[:, :, 1] - rgb[:, :, 2]) <= tolerance
  result = ~(rg_diff & rb_diff & gb_diff)

  if output_type == "bool":
    pass
  elif output_type == "float":
    result = result.astype(float)
  else:
    result = result.astype("uint8") * 255
  util.np_info(result, "Filter Grays", t.elapsed())
  return result


def uint8_to_bool(np_img):
  """
  Convert NumPy array of uint8 (255,0) values to bool (True,False) values

  Args:
    np_img: Binary image as NumPy array of uint8 (255,0) values.

  Returns:
    NumPy array of bool (True,False) values.
  """
  result = (np_img / 255).astype(bool)
  return result


def apply_image_filters(np_img, slide_num=None, info=None, save=False, display=False):
  """
  Apply filters to image as NumPy array and optionally save and/or display filtered images.

  Args:
    np_img: Image as NumPy array.
    slide_num: The slide number (used for saving/displaying).
    info: Dictionary of slide information (used for HTML display).
    save: If True, save image.
    display: If True, display image.

  Returns:
    Resulting filtered image as a NumPy array.
  """
  rgb = np_img
  save_display(save, display, info, rgb, slide_num, 1, "Original", "rgb")

  mask_not_green = filter_green_channel(rgb)
  rgb_not_green = util.mask_rgb(rgb, mask_not_green)
  save_display(save, display, info, rgb_not_green, slide_num, 2, "Not Green", "rgb-not-green")

  mask_not_gray = filter_grays(rgb)
  rgb_not_gray = util.mask_rgb(rgb, mask_not_gray)
  save_display(save, display, info, rgb_not_gray, slide_num, 3, "Not Gray", "rgb-not-gray")

  mask_no_red_pen = filter_red_pen(rgb)
  rgb_no_red_pen = util.mask_rgb(rgb, mask_no_red_pen)
  save_display(save, display, info, rgb_no_red_pen, slide_num, 4, "No Red Pen", "rgb-no-red-pen")

  mask_no_green_pen = filter_green_pen(rgb)
  rgb_no_green_pen = util.mask_rgb(rgb, mask_no_green_pen)
  save_display(save, display, info, rgb_no_green_pen, slide_num, 5, "No Green Pen", "rgb-no-green-pen")

  mask_no_blue_pen = filter_blue_pen(rgb)
  rgb_no_blue_pen = util.mask_rgb(rgb, mask_no_blue_pen)
  save_display(save, display, info, rgb_no_blue_pen, slide_num, 6, "No Blue Pen", "rgb-no-blue-pen")

  mask_gray_green_pens = mask_not_gray & mask_not_green & mask_no_red_pen & mask_no_green_pen & mask_no_blue_pen
  rgb_gray_green_pens = util.mask_rgb(rgb, mask_gray_green_pens)
  save_display(save, display, info, rgb_gray_green_pens, slide_num, 7, "Not Gray, Not Green, No Pens",
               "rgb-no-gray-no-green-no-pens")

  mask_remove_small = filter_remove_small_objects(mask_gray_green_pens, min_size=500, output_type="bool")
  rgb_remove_small = util.mask_rgb(rgb, mask_remove_small)
  save_display(save, display, info, rgb_remove_small, slide_num, 8,
               "Not Gray, Not Green, No Pens,\nRemove Small Objects",
               "rgb-not-green-not-gray-no-pens-remove-small")

  img = rgb_remove_small
  return img


def apply_filters_to_image(slide_num, save=True, display=False):
  """
  Apply a set of filters to an image and optionally save and/or display filtered images.

  Args:
    slide_num: The slide number.
    save: If True, save filtered images.
    display: If True, display filtered images to screen.

  Returns:
    Tuple consisting of 1) the resulting filtered image as a NumPy array, and 2) dictionary of image information
    (used for HTML page generation).
  """
  t = Time()
  print("Processing slide #%d" % slide_num)

  info = dict()

  if save and not os.path.exists(slide.FILTER_DIR):
    os.makedirs(slide.FILTER_DIR)
  img_path = slide.get_training_image_path(slide_num)
  np_orig = slide.open_image_np(img_path)
  filtered_np_img = apply_image_filters(np_orig, slide_num, info, save=save, display=display)

  if save:
    t1 = Time()
    result_path = slide.get_filter_image_result(slide_num)
    pil_img = util.np_to_pil(filtered_np_img)
    pil_img.save(result_path)
    print("%-20s | Time: %-14s  Name: %s" % ("Save Image", str(t1.elapsed()), result_path))

    t1 = Time()
    thumbnail_path = slide.get_filter_thumbnail_result(slide_num)
    slide.save_thumbnail(pil_img, slide.THUMBNAIL_SIZE, thumbnail_path)
    print("%-20s | Time: %-14s  Name: %s" % ("Save Thumbnail", str(t1.elapsed()), thumbnail_path))

  print("Slide #%03d processing time: %s\n" % (slide_num, str(t.elapsed())))

  return filtered_np_img, info


def save_display(save, display, info, np_img, slide_num, filter_num, display_text, file_text,
                 display_mask_percentage=True):
  """
  Optionally save an image and/or display the image.

  Args:
    save: If True, save filtered images.
    display: If True, display filtered images to screen.
    info: Dictionary to store filter information.
    np_img: Image as a NumPy array.
    slide_num: The slide number.
    filter_num: The filter number.
    display_text: Filter display name.
    file_text: Filter name for file.
    display_mask_percentage: If True, display mask percentage on displayed slide.
  """
  mask_percentage = None
  if display_mask_percentage:
    mask_percentage = mask_percent(np_img)
    display_text = display_text + "\n(" + mask_percentage_text(mask_percentage) + " masked)"
  if slide_num is None and filter_num is None:
    pass
  elif filter_num is None:
    display_text = "S%03d " % slide_num + display_text
  elif slide_num is None:
    display_text = "F%03d " % filter_num + display_text
  else:
    display_text = "S%03d-F%03d " % (slide_num, filter_num) + display_text
  if display:
    util.display_img(np_img, display_text)
  if save:
    save_filtered_image(np_img, slide_num, filter_num, file_text)
  if info is not None:
    info[slide_num * 1000 + filter_num] = (slide_num, filter_num, display_text, file_text, mask_percentage)


def mask_percentage_text(mask_percentage):
  """
  Generate a formatted string representing the percentage that an image is masked.

  Args:
    mask_percentage: The mask percentage.

  Returns:
    The mask percentage formatted as a string.
  """
  return "%3.2f%%" % mask_percentage


def image_cell(slide_num, filter_num, display_text, file_text):
  """
  Generate HTML for viewing a processed image.

  Args:
    slide_num: The slide number.
    filter_num: The filter number.
    display_text: Filter display name.
    file_text: Filter name for file.

  Returns:
    HTML for a table cell for viewing a filtered image.
  """
  filt_img = slide.get_filter_image_path(slide_num, filter_num, file_text)
  filt_thumb = slide.get_filter_thumbnail_path(slide_num, filter_num, file_text)
  img_name = slide.get_filter_image_filename(slide_num, filter_num, file_text)
  return "      <td>\n" + \
         "        <a target=\"_blank\" href=\"%s\">%s<br/>\n" % (filt_img, display_text) + \
         "          <img src=\"%s\" />\n" % (filt_thumb) + \
         "        </a>\n" + \
         "      </td>\n"


def html_header(page_title):
  """
  Generate an HTML header for previewing images.

  Returns:
    HTML header for viewing images.
  """
  html = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" " + \
         "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n" + \
         "<html xmlns=\"http://www.w3.org/1999/xhtml\" lang=\"en\" xml:lang=\"en\">\n" + \
         "  <head>\n" + \
         "    <title>%s</title>\n" % page_title + \
         "    <style type=\"text/css\">\n" + \
         "     img { border: 2px solid black; }\n" + \
         "     td { border: 2px solid black; }\n" + \
         "    </style>\n" + \
         "  </head>\n" + \
         "  <body>\n"
  return html


def html_footer():
  """
  Generate an HTML footer for previewing images.

  Returns:
    HTML footer for viewing images.
  """
  html = "</body>\n" + \
         "</html>\n"
  return html


def save_filtered_image(np_img, slide_num, filter_num, filter_text):
  """
  Save a filtered image to the file system.

  Args:
    np_img: Image as a NumPy array.
    slide_num:  The slide number.
    filter_num: The filter number.
    filter_text: Descriptive text to add to the image filename.
  """
  t = Time()
  filepath = slide.get_filter_image_path(slide_num, filter_num, filter_text)
  pil_img = util.np_to_pil(np_img)
  pil_img.save(filepath)
  print("%-20s | Time: %-14s  Name: %s" % ("Save Image", str(t.elapsed()), filepath))

  t1 = Time()
  thumbnail_filepath = slide.get_filter_thumbnail_path(slide_num, filter_num, filter_text)
  slide.save_thumbnail(pil_img, slide.THUMBNAIL_SIZE, thumbnail_filepath)
  print("%-20s | Time: %-14s  Name: %s" % ("Save Thumbnail", str(t1.elapsed()), thumbnail_filepath))


def generate_filter_html_result(html_page_info):
  """
  Generate HTML to view the filtered images. If slide.FILTER_PAGINATE is True, the results will be paginated.

  Args:
    html_page_info: Dictionary of image information.
  """
  if not slide.FILTER_PAGINATE:
    html = ""
    html += html_header("Filtered Images")
    html += "  <table>\n"

    row = 0
    for key in sorted(html_page_info):
      value = html_page_info[key]
      current_row = value[0]
      if current_row > row:
        html += "    <tr>\n"
        row = current_row
      html += image_cell(value[0], value[1], value[2], value[3])
      next_key = key + 1
      if next_key not in html_page_info:
        html += "    </tr>\n"

    html += "  </table>\n"
    html += html_footer()
    text_file = open(os.path.join(slide.FILTER_HTML_DIR, "filters.html"), "w")
    text_file.write(html)
    text_file.close()
  else:
    slide_nums = set()
    for key in html_page_info:
      slide_num = math.floor(key / 1000)
      slide_nums.add(slide_num)
    slide_nums = sorted(list(slide_nums))
    total_len = len(slide_nums)
    page_size = slide.FILTER_PAGINATION_SIZE
    num_pages = math.ceil(total_len / page_size)

    for page_num in range(1, num_pages + 1):
      start_index = (page_num - 1) * page_size
      end_index = (page_num * page_size) if (page_num < num_pages) else total_len
      page_slide_nums = slide_nums[start_index:end_index]

      html = ""
      html += html_header("Filtered Images, Page %d" % page_num)

      html += "  <div style=\"font-size: 20px\">"
      if page_num > 1:
        if page_num == 2:
          html += "<a href=\"filters.html\">&lt;</a> "
        else:
          html += "<a href=\"filters-%d.html\">&lt;</a> " % (page_num - 1)
      html += "Page %d" % page_num
      if page_num < num_pages:
        html += " <a href=\"filters-%d.html\">&gt;</a> " % (page_num + 1)
      html += "</div>\n"

      html += "  <table>\n"
      for slide_num in page_slide_nums:
        html += "  <tr>\n"
        filter_num = 1

        lookup_key = slide_num * 1000 + filter_num
        while lookup_key in html_page_info:
          value = html_page_info[lookup_key]
          html += image_cell(value[0], value[1], value[2], value[3])
          lookup_key += 1
        html += "  </tr>\n"

      html += "  </table>\n"

      html += html_footer()
      if page_num == 1:
        text_file = open(os.path.join(slide.FILTER_HTML_DIR, "filters.html"), "w")
      else:
        text_file = open(os.path.join(slide.FILTER_HTML_DIR, "filters-%d.html" % page_num), "w")
      text_file.write(html)
      text_file.close()


def apply_filters_to_image_list(image_num_list, save, display):
  """
  Apply filters to a list of images.

  Args:
    image_num_list: List of image numbers.
    save: If True, save filtered images.
    display: If True, display filtered images to screen.

  Returns:
    Tuple consisting of 1) a list of image numbers, and 2) a dictionary of image filter information.
  """
  html_page_info = dict()
  for slide_num in image_num_list:
    _, info = apply_filters_to_image(slide_num, save=save, display=display)
    html_page_info.update(info)
  return image_num_list, html_page_info


def apply_filters_to_image_range(start_ind, end_ind, save, display):
  """
  Apply filters to a range of images.

  Args:
    start_ind: Starting index (inclusive).
    end_ind: Ending index (inclusive).
    save: If True, save filtered images.
    display: If True, display filtered images to screen.

  Returns:
    Tuple consisting of 1) staring index of slides converted to images, 2) ending index of slides converted to images,
    and 3) a dictionary of image filter information.
  """
  html_page_info = dict()
  for slide_num in range(start_ind, end_ind + 1):
    _, info = apply_filters_to_image(slide_num, save=save, display=display)
    html_page_info.update(info)
  return start_ind, end_ind, html_page_info


def singleprocess_apply_filters_to_images(save=True, display=False, html=True, image_num_list=None):
  """
  Apply a set of filters to training images and optionally save and/or display the filtered images.

  Args:
    save: If True, save filtered images.
    display: If True, display filtered images to screen.
    html: If True, generate HTML page to display filtered images.
    image_num_list: Optionally specify a list of image slide numbers.
  """
  t = Time()
  print("Applying filters to images\n")

  if image_num_list is not None:
    _, info = apply_filters_to_image_list(image_num_list, save, display)
  else:
    num_training_slides = slide.get_num_training_slides()
    (s, e, info) = apply_filters_to_image_range(1, num_training_slides, save, display)

  print("Time to apply filters to all images: %s\n" % str(t.elapsed()))

  if html:
    generate_filter_html_result(info)


def multiprocess_apply_filters_to_images(save=True, display=False, html=True, image_num_list=None):
  """
  Apply a set of filters to all training images using multiple processes (one process per core).

  Args:
    save: If True, save filtered images.
    display: If True, display filtered images to screen (multiprocessed display not recommended).
    html: If True, generate HTML page to display filtered images.
    image_num_list: Optionally specify a list of image slide numbers.
  """
  timer = Time()
  print("Applying filters to images (multiprocess)\n")

  if save and not os.path.exists(slide.FILTER_DIR):
    os.makedirs(slide.FILTER_DIR)

  # how many processes to use
  num_processes = multiprocessing.cpu_count()
  pool = multiprocessing.Pool(num_processes)

  if image_num_list is not None:
    num_train_images = len(image_num_list)
  else:
    num_train_images = slide.get_num_training_slides()
  if num_processes > num_train_images:
    num_processes = num_train_images
  images_per_process = num_train_images / num_processes

  print("Number of processes: " + str(num_processes))
  print("Number of training images: " + str(num_train_images))

  tasks = []
  for num_process in range(1, num_processes + 1):
    start_index = (num_process - 1) * images_per_process + 1
    end_index = num_process * images_per_process
    start_index = int(start_index)
    end_index = int(end_index)
    if image_num_list is not None:
      sublist = image_num_list[start_index - 1:end_index]
      tasks.append((sublist, save, display))
      print("Task #" + str(num_process) + ": Process slides " + str(sublist))
    else:
      tasks.append((start_index, end_index, save, display))
      if start_index == end_index:
        print("Task #" + str(num_process) + ": Process slide " + str(start_index))
      else:
        print("Task #" + str(num_process) + ": Process slides " + str(start_index) + " to " + str(end_index))

  # start tasks
  results = []
  for t in tasks:
    if image_num_list is not None:
      results.append(pool.apply_async(apply_filters_to_image_list, t))
    else:
      results.append(pool.apply_async(apply_filters_to_image_range, t))

  html_page_info = dict()
  for result in results:
    if image_num_list is not None:
      (image_nums, html_page_info_res) = result.get()
      html_page_info.update(html_page_info_res)
      print("Done filtering slides: %s" % image_nums)
    else:
      (start_ind, end_ind, html_page_info_res) = result.get()
      html_page_info.update(html_page_info_res)
      if (start_ind == end_ind):
        print("Done filtering slide %d" % start_ind)
      else:
        print("Done filtering slides %d through %d" % (start_ind, end_ind))

  if html:
    generate_filter_html_result(html_page_info)

  print("Time to apply filters to all images (multiprocess): %s\n" % str(timer.elapsed()))

# if __name__ == "__main__":
# slide.training_slide_to_image(2)
# singleprocess_apply_filters_to_images(image_num_list=[2], display=True)

# singleprocess_apply_filters_to_images()
# multiprocess_apply_filters_to_images()



"""Image Writing 
==========
Contains functions for stitching images together."""
import pandas as pd, numpy as np
from scipy.ndimage import label as scilabel
from skimage.measure import regionprops_table
import cv2, os, subprocess
from deepzoom import *
from deepzoom import _get_or_create_path,_get_files_path
from PIL import Image
import tqdm
import dask
from dask.diagnostics import ProgressBar
from scipy.special import softmax
import torch
from sauth import SimpleHTTPAuthHandler, serve_http
from skimage.draw import circle
from pathpretrain.utils import load_image
Image.MAX_IMAGE_PIXELS = None

class Numpy2DZI(ImageCreator):
    def __init__(
        self,
        tile_size=254,
        tile_overlap=1,
        tile_format="jpg",
        image_quality=0.8,
        resize_filter=None,
        copy_metadata=False,
        compression=1.
    ):
        super().__init__(tile_size,tile_overlap,tile_format,image_quality,resize_filter,copy_metadata)
        self.compression=compression

    def create(self, source_arr, destination):
        """
        Create a deep zoom image from the given source array.

        Parameters
        ----------
        source_arr : ndarray
            The source image as a NumPy array.
        destination : str
            The destination folder to save the tiles to.

        Returns
        -------
        str
            The destination folder where the tiles were saved.
        """
        # potentially have an option where dynamically softlink once deeper layer is made so slide is readily available, push to background process and write metadata for dash app to read
        # speed up image saving with dask https://stackoverflow.com/questions/54615625/how-to-save-dask-array-as-png-files-slice-by-slice https://github.com/dask/dask-image/issues/110
        self.image = PIL.Image.fromarray(source_arr if self.compression==1 else cv2.resize(source_arr,None,fx=1/self.compression,fy=1/self.compression,interpolation=cv2.INTER_CUBIC))
        width, height = self.image.size
        self.descriptor = DeepZoomImageDescriptor(
            width=width,
            height=height,
            tile_size=self.tile_size,
            tile_overlap=self.tile_overlap,
            tile_format=self.tile_format,
        )
        image_files = _get_or_create_path(_get_files_path(destination))
        for level in tqdm.trange(self.descriptor.num_levels, desc='level'):
            level_dir = _get_or_create_path(os.path.join(image_files, str(level)))
            level_image = self.get_image(level)
            for (column, row) in tqdm.tqdm(self.tiles(level), desc='tiles'):
                bounds = self.descriptor.get_tile_bounds(level, column, row)
                tile = level_image.crop(bounds)
                format = self.descriptor.tile_format
                tile_path = os.path.join(level_dir, "%s_%s.%s" % (column, row, format))
                tile_file = open(tile_path, "wb")
                if self.descriptor.tile_format == "jpg":
                    jpeg_quality = int(self.image_quality * 100)
                    tile.save(tile_file, "JPEG", quality=jpeg_quality)
                else:
                    tile.save(tile_file)
        self.descriptor.save(destination)
        return destination

def stitch_slide(arr,
                 dzi_out,
                 compression=4,
                 mask=None):
    if mask is not None: arr[mask]=255
    Numpy2DZI(compression=compression).create(arr,
                                              dzi_out)

def npy2dzi(npy_file='',
            dzi_out='',
            compression=1.):
    stitch_slide(load_image(npy_file),compression=compression,dzi_out=dzi_out)


def stitch_slides(basename="",
                 compression=4,
                 dirname=".",
                 ext=".tif",
                 pull_mask=False,
                 mask_compresssion=8):
    image=os.path.join(dirname,"inputs",f"{basename}{ext}")
    basename=os.path.basename(image).replace(ext,'')
    image=load_image(image)#np.load(image)
    xy_bounds=pd.read_pickle(os.path.join(dirname,"masks",f"{basename}.pkl"))
    write_files=[]
    for ID in xy_bounds:
        (xmin,ymin),(xmax,ymax)=xy_bounds[ID]
        dzi_out=os.path.join(dirname,"dzi_files",f"{basename}_{ID}_img.dzi")
        arr=image[xmin:xmax,ymin:ymax]
        mask=None
        if pull_mask:
            mask=dask.delayed(lambda x: cv2.resize(np.load(x).astype(int),arr.shape[:2],interpolation=cv2.INTER_NEAREST).astype(bool))(os.path.join(dirname,"masks",f"{basename}_{ID}.npy"))
        write_files.append(dask.delayed(stitch_slide)(arr,dzi_out,compression,mask))
    with ProgressBar():
        dask.compute(write_files,scheduler='threading')



"""
Tumor Mapping
==========
Contains functions for creating tumor maps at surgical site. 
Will contain functions for optimal transport mapping. Code is currently being refactored and this section will be updated."""
# TODO



"""Serial Workflow 
==========
Contains functions for serial processing of tissue sections. 
Contains functions for defining and running workflows."""
import glob, os, time, pickle
from .preprocessing import preprocess
from .cnn_prediction import generate_embeddings
from .generate_graph import create_graph_data
from .gnn_prediction import predict
# from .quality_scores import generate_quality_scores
from .ink_detection import detect_inks
#from .compile_results import dump_results
#from .nuclei_detection import predict_nuclei
from .image_stitch import npy2dzi, stitch_slides
#from .case_prototype import Case
import warnings
from functools import partial
import pandas as pd

def files_exist_overwrite(overwrite, files):
    return (not overwrite) and all([os.path.exists(file) for file in files])

def generate_output_file_names(basename):
    out_files=dict()
    out_files['preprocess']=[f"masks/{basename}_{k}_map.npy" for k in ['tumor','macro']]+[f"patches/{basename}_{k}_map.npy" for k in ['tumor','macro']]
    for k in ['macro','tumor']:
        out_files[f'cnn_{k}']=[f"cnn_embeddings/{basename}_{k}_map.pkl"]
        out_files[f'graph_data_{k}']=[f"graph_datasets/{basename}_{k}_map.pkl"]
        out_files[f'gnn_{k}']=[f"gnn_results/{basename}_{k}_map.pkl"]
    out_files['quality']=[f"quality_scores/{basename}.pkl"]
    out_files['ink']=[f"detected_inks/{basename}_thumbnail.npy"]
    out_files['nuclei']=[f"nuclei_results/{basename}.npy"]
    return out_files

def run_workflow_series(basename, compression, overwrite, ext, dirname, df_section_pieces_file, run_stitch_slide):
    """
    Runs image processing workflow in series on an input image.

    Parameters
    ----------
    basename : str
        The base name of the slide to process.
    compression : float
        The level of compression to apply to the slide.
    overwrite : bool
        Whether to overwrite existing files if they exist.
    ext : str
        The file extension of the slide.
    dirname : str
        The directory containing the slide and other relevant files.
    df_section_pieces_file : str
        The file containing information about the patches.
    run_stitch_slide : bool
        Whether to run the stitch_slides function after all other processing is complete.

    Returns
    -------
    times : dict
        A dictionary containing the times at which each step of the workflow was completed.
    """

    print(f"{basename} preprocessing")

    out_files=generate_output_file_names(basename)

    times=dict()
    times['start']=time.time()

    if not files_exist_overwrite(overwrite,out_files['preprocess']):
        preprocess(basename=basename,
               threshold=0.05,
               patch_size=256,
               ext=ext,
               no_break=False,
               df_section_pieces_file=df_section_pieces_file,
               image_mask_compression=8.,
               dirname=dirname)

    times['preprocess']=time.time()

    new_basenames=pd.read_pickle(os.path.join(dirname,"metadata",f"{basename}.pkl"))

    for k in ['tumor','macro']: times[f'cnn_{k}'],times[f'graph_{k}'],times[f'gnn_{k}']=[],[],[]
    for bn in new_basenames:
        for k in ['tumor','macro']:
            print(f"{bn} {k} embedding")
            if not files_exist_overwrite(overwrite,out_files[f'cnn_{k}']):
                generate_embeddings(basename=bn,
                                analysis_type=k,
                               gpu_id=-1,
                               dirname=dirname)
            times[f'cnn_{k}'].append(time.time())

            print(f"{bn} {k} build graph")
            if not files_exist_overwrite(overwrite,out_files[f'graph_data_{k}']):
                create_graph_data(basename=bn,
                              analysis_type=k,
                              radius=256,
                              min_component_size=600,
                              no_component_break=True,
                              dirname=dirname)
            times[f'graph_{k}'].append(time.time())

            print(f"{bn} {k} gnn predict")
            if not files_exist_overwrite(overwrite,out_files[f'gnn_{k}']):
                predict(basename=bn,
                    analysis_type=k,
                    gpu_id=-1,
                    dirname=dirname)
            times[f'gnn_{k}'].append(time.time())

    print(f"{basename} ink detection")
    if not files_exist_overwrite(overwrite,out_files['ink']):
        detect_inks(basename=basename,
                compression=8.,
                ext=ext,
                dirname=dirname)
    times["ink"]=time.time()

    print(f"{basename} stitch")
    if not files_exist_overwrite(overwrite,out_files['ink']):
        if run_stitch_slide:
            stitch_slides(basename=basename,
                compression=4,
                ext=ext,
                dirname=dirname)
            times["stitch"]=time.time()

    return times


def run_series(patient="163_A1",
               input_dir="inputs",
               compression=1.,
               overwrite=True,
               record_time=False,
               ext=".npy",
               dirname=".",
               df_section_pieces_file="df_section_pieces.pkl",
               run_stitch_slide=True
               ):
    for f in glob.glob(os.path.join(input_dir,f"{patient}*{ext}")):
        basename=os.path.basename(f).replace(ext,"")#.replace(".tiff","").replace(".tif","").replace(".svs","")
        times=run_workflow_series(basename,
                            compression,
                            overwrite,
                            ext,
                            dirname,
                            df_section_pieces_file,
                            run_stitch_slide)
        if record_time:
            os.makedirs(os.path.join(dirname,"times"),exist_ok=True)
            pickle.dump(times,open(os.path.join(dirname,"times",f"{basename}.pkl"),'wb'))



"""Parallel Workflow 
==========
Contains functions for parallel processing of tissue sections."""
from toil.job import Job
import subprocess, os
import glob
import fire

def files_exist_overwrite(overwrite, files):
    return (not overwrite) and all([os.path.exists(file) for file in files])

def generate_output_file_names(basename):
    out_files=dict()
    out_files['preprocess']=[f"masks/{basename}_{k}_map.npy" for k in ['tumor','macro']]+[f"patches/{basename}_{k}_map.npy" for k in ['tumor','macro']]
    for k in ['macro','tumor']:
        out_files[f'cnn_{k}']=[f"cnn_embeddings/{basename}_{k}_map.pkl"]
        out_files[f'gnn_{k}']=[f"gnn_results/{basename}_{k}_map.pkl",f"graph_datasets/{basename}_{k}_map.pkl"]
    out_files['quality']=[f"quality_scores/{basename}.pkl"]
    out_files['ink']=[f"detected_inks/{basename}_thumbnail.npy"]
    out_files['nuclei']=[f"nuclei_results/{basename}.npy"]
    return out_files

def preprocess(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai preprocess --basename {job_dict['basename']} --threshold 0.05 --patch_size 256 --ext {job_dict['ext']} --compression {job_dict['compression']}"
    # print(command)
    result=os.popen(command).read()
    return result#f"Preprocessed {job_dict['basename']}"

def embed(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai cnn_predict --basename {job_dict['basename']} --analysis_type {job_dict['analysis_type']} --gpu_id -1"
    # print(command)
    result=os.popen(command).read()
    return result#f"Embed CNN {job_dict['analysis_type']} {job_dict['basename']}"

def gnn_predict(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai gnn_predict --basename {job_dict['basename']} --analysis_type {job_dict['analysis_type']} --radius 256 --min_component_size 600 --gpu_id -1 --generate_graph True"
    # print(command)
    result=os.popen(command).read()
    return result#f"GNN Predict {job_dict['analysis_type']} {job_dict['basename']}"

def gen_quality_scores(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai quality_score --basename {job_dict['basename']} "
    # print(command)
    result=os.popen(command).read()
    return result#f"Quality {job_dict['basename']}"

def ink_detect(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai ink_detect --basename {job_dict['basename']} --compression 8 --ext {job_dict['ext']}"
    # print(command)
    result=os.popen(command).read()
    return result#f"Ink {job_dict['basename']}"

def stitch_images(job, job_dict, memory="50G", cores=8, disk="1M"):
    command=f"cd {job_dict['job_dir']} && {job_dict['singularity_preamble']} arctic_ai ink_detect --basename {job_dict['basename']} --compression 8 --ext {job_dict['ext']}"
    # print(command)
    result=os.popen(command).read()
    return result#f"Ink {job_dict['basename']}"

def deploy_patient(job, job_dict, memory="2G", cores=2, disk="1M"):
    os.chdir(job_dict['job_dir'])
    out_files=generate_output_file_names(job_dict['basename'])
    
    jobs={}
    preprocess_job=job.addChildJobFn(preprocess, job_dict, memory, cores, disk)
    jobs['preprocess']=preprocess_job
    
    embed_jobs={}
    gnn_predict_jobs={}
    for k in ['tumor','macro']:
        job_dict_k=job_dict.copy()
        job_dict_k['analysis_type']=k
        embed_jobs[k]=job.addChildJobFn(embed, job_dict_k, memory, cores, disk)
        gnn_predict_jobs[k]=job.addChildJobFn(gnn_predict, job_dict_k, memory, cores, disk)
        jobs['preprocess'].addChild(embed_jobs[k])
        embed_jobs[k].addChild(gnn_predict_jobs[k])
    jobs['embed']=embed_jobs
    jobs['gnn']=gnn_predict_jobs
    # quality_job=job.addChildJobFn(gen_quality_scores, job_dict, memory, cores, disk)
    # for k in ['tumor','macro']:
    #     gnn_predict_jobs[k].addChild(quality_job)
    ink_job=job.addChildJobFn(ink_detect, job_dict, memory, cores, disk)
    jobs['preprocess'].addChild(ink_job)
    jobs['preprocess'].addChild(nuclei_job)
    
    return f"Processed {job_dict['basename']}"

def setup_deploy(job, job_dict, memory="2G", cores=2, disk="3G"):
    os.chdir(job_dict['job_dir'])
    jobs=[]
    for f in glob.glob(os.path.join(job_dict['input_dir'],f"{job_dict['patient']}*{job_dict['ext']}")):
        # print(f)
        basename=os.path.basename(f).replace(job_dict['ext'],"")
        job_dict_f=dict(basename=basename, 
                        compression=job_dict['compression'], 
                        overwrite=job_dict['overwrite'], 
                        ext=job_dict['ext'],
                        job_dir=job_dict['job_dir'],
                        singularity_preamble=job_dict['singularity_preamble'])
        patient_job=job.addChildJobFn(deploy_patient, job_dict_f, memory, cores, disk)
        jobs.append(patient_job)
    return [patient_job.rv() for patient_job in jobs]
                    

def run_parallel(patient="",
               input_dir="inputs",
               scheme="2/1",
               compression=6.,
               overwrite=True,
               record_time=False,
               extract_dzi=False,
               ext=".tif",
               job_dir="./",
               restart=False,
               logfile="",
               loglevel="",
               cuda_visible_devices="$CUDA_VISIBLE_DEVICES",
               singularity_img_path="arcticai.img",
               run_slurm=False,
               cores=2,
               memory="60G",
               disk="3G",
               cuda_device="$(($RANDOM % 4))",
               prepend_path="$(realpath ~)/.local/bin/",
               slurm_gpus=1,
               slurm_account="qdp-alpha",
               slurm_partition="v100_12",
               time=1,
               gpu_cmode="exclusive"):
    """
    Runs the image processing workflow in parallel across multiple tissue sections simultaneously.

    Parameters:
    -----------
    patient : str, optional
        The patient identifier. Default is an empty string.
    input_dir : str, optional
        The directory containing the input images. Default is 'inputs'.
    scheme : str, optional
        The grid scheme used for image processing. Default is '2/1'.
    compression : float, optional
        The compression level for the output images. Default is 6.0.
    overwrite : bool, optional
        Whether or not to overwrite existing output files. Default is True.
    record_time : bool, optional
        Whether or not to record the processing time for each image. Default is False.
    extract_dzi : bool, optional
        Whether or not to extract deep zoom images. Default is False.
    ext : str, optional
        The file extension for input images. Default is '.tif'.
    job_dir : str, optional
        The directory for job management. Default is './'.
    restart : bool, optional
        Whether or not to restart a previous job. Default is False.
    logfile : str, optional
        The path to the log file. Default is an empty string.
    loglevel : str, optional
        The level of verbosity for logging. Default is an empty string.
    cuda_visible_devices : str, optional
        The value for the CUDA_VISIBLE_DEVICES environment variable. Default is '$CUDA_VISIBLE_DEVICES'.
    singularity_img_path : str, optional
        The path to the Singularity image file. Default is 'arcticai.img'.
    run_slurm : bool, optional
        Whether or not to run the job using the Slurm scheduler. Default is False.
    cores : int, optional
        The number of CPU cores to use. Default is 2.
    memory : str, optional
        The amount of memory to allocate for each job. Default is '60G'.
    disk : str, optional
        The amount of disk space to allocate for each job. Default is '3G'.
    cuda_device : str, optional
        The value for the CUDA device ID. Default is '$(($RANDOM % 4))'.
    prepend_path : str, optional
        The value for the PREPEND_PATH environment variable. Default is '$(realpath ~)/.local/bin/'.
    slurm_gpus : int, optional
        The number of GPUs to allocate when running the job on a Slurm cluster. Default is 1.
    slurm_account : str, optional
        The name of the Slurm account to use. Default is 'qdp-alpha'.
    slurm_partition : str, optional
        The name of the Slurm partition to use. Default is 'v100_12'.
    time : int, optional
        The maximum amount of time to allocate for the job, in hours. Default is 1.
    gpu_cmode : str, optional
        Slurm option that controls GPU sharing mode between multiple users. Default is exclusive though can be toggled to shared
    """
    singularity_preamble=f"source ~/.bashrc && export SINGULARITYENV_CUDA_VISIBLE_DEVICES={cuda_visible_devices} && export SINGULARITYENV_PREPEND_PATH={prepend_path} && singularity  exec --nv -B $(pwd)  -B /scratch/  --bind ${HOME}:/mnt  {singularity_img_path}"
    slurm_args=f"--export=ALL --gres=gpu:{slurm_gpus} --account={slurm_account} --partition={slurm_partition} --nodes=1 --ntasks-per-node=1 --time={time}:00:00 "
    if gpu_cmode=="shared":
        slurm_args+="--gpu_cmode=shared"
    options = Job.Runner.getDefaultOptions("./toilWorkflowRun")
    options.restart=restart
    options.defaultCores=cores
    options.defaultMemory=memory
    options.defaultDisk=disk
    options.clean = "always"
    if run_slurm: 
        os.environ["TOIL_SLURM_ARGS"]=slurm_args
        options.batchSystem="slurm"
        options.disableCaching=True
        options.statePollingWait = 5
        options.maxLocalJobs = 100
        options.targetTime = 1
    else:
        singularity_preamble=f"export CUDA_VISIBLE_DEVICES={cuda_device} &&"+singularity_preamble
    if loglevel: options.logLevel=loglevel
    if logfile: options.logFile=logfile
    job_dict=dict(patient=patient,
               input_dir=input_dir,
               scheme=scheme,
               compression=compression,
               overwrite=overwrite,
               record_time=record_time,
               extract_dzi=extract_dzi,
               ext=ext,
               job_dir=job_dir,
               singularity_preamble=singularity_preamble)
    j = Job.wrapJobFn(setup_deploy, job_dict)
    rv = Job.Runner.startToil(j, options)

# if __name__=="__main__":
#     fire.Fire(run_parallel)


"""
Graph Data 
==========
Functions for graph dataset generation.
"""
import os, torch, numpy as np, pandas as pd
import pickle
import scipy.sparse as sps
from torch_geometric.utils import subgraph, add_remaining_self_loops
from torch_cluster import radius_graph
from collections import Counter
from torch_geometric.data import Data

def create_graph_data(basename="163_A1a",
                      analysis_type="tumor",
                      radius=256,
                      min_component_size=600,
                      no_component_break=False,
                      dirname="."):
    """
    Creates graph data for use in the GNN model for a given tissue slide.

    Parameters
    ----------
    basename : str
        The basename of the tissue slide to create graph data for.
    analysis_type : str
        The type of analysis to perform. Can be "tumor" or "macro".
    radius : int
        The radius to use when creating the graph.
    min_component_size : int
        The minimum size a connected component must be to be included in the graph data.
    no_component_break : bool
        Whether to include all connected components in the graph data, or just the largest one.
    dirname : str
        The directory to save the graph data in.

    Returns
    -------
    None
    """

    os.makedirs(os.path.join(dirname,"graph_datasets"),exist_ok=True)

    embeddings=torch.load(os.path.join(dirname,f"cnn_embeddings/{basename}_{analysis_type}_map.pkl"))
    xy=torch.tensor(embeddings['patch_info'][['x','y']].values).float()
    if torch.cuda.is_available():
        xy=xy.cuda()
    X=torch.tensor(embeddings['embeddings'])
    G=radius_graph(xy, r=radius*np.sqrt(2), batch=None, loop=True)
    G=G.detach().cpu()
    G=add_remaining_self_loops(G)[0]
    xy=xy.detach().cpu()
    datasets=[]
    edges=G.detach().cpu().numpy().astype(int)
    n_components,components=list(sps.csgraph.connected_components(sps.coo_matrix((np.ones_like(edges[0]),(edges[0],edges[1])))))
    comp_count=Counter(components)
    components=torch.LongTensor(components)
    if not no_component_break:
        for i in range(n_components):
            if comp_count[i]>=min_component_size:
                G_new=subgraph(components==i,G,relabel_nodes=True)[0]
                xy_new=xy[components==i]
                X_new=X[components==i]
                np.random.seed(42)
                idx=np.arange(X_new.shape[0])
                idx2=np.arange(X_new.shape[0])
                np.random.shuffle(idx)
                train_idx,val_idx,test_idx=torch.tensor(np.isin(idx2,idx[:int(0.8*len(idx))])),torch.tensor(np.isin(idx2,idx[int(0.8*len(idx)):int(0.9*len(idx))])),torch.tensor(np.isin(idx2,idx[int(0.9*len(idx)):]))
                dataset=Data(x=X_new, edge_index=G_new, y_new=torch.ones(len(X_new)), edge_attr=None, pos=xy_new)
                dataset.train_mask=train_idx
                dataset.val_mask=val_idx
                dataset.test_mask=test_idx
                dataset.id=basename
                dataset.component=i
                datasets.append(dataset)
    else:
        dataset=Data(x=X, edge_index=G, y_new=torch.ones(len(X)), edge_attr=None, pos=xy)
        np.random.seed(42)
        idx=np.arange(X.shape[0])
        idx2=np.arange(X.shape[0])
        np.random.shuffle(idx)
        train_idx,val_idx,test_idx=torch.tensor(np.isin(idx2,idx[:int(0.8*len(idx))])),torch.tensor(np.isin(idx2,idx[int(0.8*len(idx)):int(0.9*len(idx))])),torch.tensor(np.isin(idx2,idx[int(0.9*len(idx)):]))
        dataset.train_mask=train_idx
        dataset.val_mask=val_idx
        dataset.test_mask=test_idx
        dataset.id=basename
        dataset.component=0
        datasets.append(dataset)

    pickle.dump(datasets,open(os.path.join(dirname,'graph_datasets',f"{basename}_{analysis_type}_map.pkl"),'wb'))



"""Preprocess 
==========
Contains functions for preprocessing images."""
import os, tqdm
import numpy as np, pandas as pd
from itertools import product
from scipy.ndimage.morphology import binary_fill_holes as fill_holes
from pathpretrain.utils import load_image, generate_tissue_mask
from scipy.sparse.csgraph import connected_components
from sklearn.neighbors import radius_neighbors_graph
from sklearn.cluster import SpectralClustering
from shapely.geometry import Point, MultiPoint
import cv2
import dask
from dask.diagnostics import ProgressBar
import alphashape
import warnings
# import pysnooper

def preprocess(basename="163_A1a",
               threshold=0.05,
               patch_size=256,
               ext='.npy',
               secondary_patch_size=0,
               alpha=1024**-1,
               no_break=False,
               df_section_pieces_file='',
               dirname=".",
               image_mask_compression=1.,
               use_section=False
               ):
    """
    Preprocesses image and generates patches for training or testing.

    Parameters
    ----------
    basename: str
        The base filename of the image to be preprocessed.
    threshold: float, optional
        The maximum fraction of a patch that can be blank. Default is 0.05.
    patch_size: int, optional
        The size of the patches to be generated. Default is 256.
    ext: str, optional
        The file extension of the input image. Default is ".npy".
    no_break: bool, optional
        If True, the function will not break large images into multiple smaller ones. Default is False.
    df_section_pieces_file: str, optional
        The filename of the file containing metadata about image patches. Default is "section_pieces.pkl".
    image_mask_compression: float, optional
        The degree of compression applied to the image mask. Default is 8.
    dirname: str, optional
        The directory where input and output files are stored. Default is ".".
    """

    assert secondary_patch_size==0
    write_images=False
    os.makedirs(os.path.join(dirname,"masks"),exist_ok=True)
    os.makedirs(os.path.join(dirname,"patches"),exist_ok=True)
    os.makedirs(os.path.join(dirname,"images"),exist_ok=True)
    os.makedirs(os.path.join(dirname,"metadata"),exist_ok=True)

    image=os.path.join(dirname,"inputs",f"{basename}{ext}")
    basename=os.path.basename(image).replace(ext,'')
    image=load_image(image)#np.load(image)
    img_shape=image.shape[:-1]
    df_section_pieces=None if not (df_section_pieces_file and os.path.exists(df_section_pieces_file)) else pd.read_pickle(df_section_pieces_file).reset_index().drop_duplicates().set_index("index")

    masks=dict()
    masks['tumor_map']=generate_tissue_mask(image,
                             compression=10,
                             otsu=False,
                             threshold=240,
                             connectivity=8,
                             kernel=5,
                             min_object_size=100000,
                             return_convex_hull=False,
                             keep_holes=False,
                             max_hole_size=6000,
                             gray_before_close=True,
                             blur_size=51)

    x_max,y_max=masks['tumor_map'].shape
    if no_break: masks['macro_map']=fill_holes(masks['tumor_map'])

    patch_info=dict()
    patches=dict()
    include_patches=dict()
    patch_info['orig']=pd.DataFrame([[basename,x,y,patch_size,"0"] for x,y in tqdm.tqdm(list(product(range(0,x_max-patch_size,patch_size),range(0,y_max-patch_size,patch_size))))],columns=['ID','x','y','patch_size','annotation'])

    for k in (masks if no_break else ['tumor_map']):
        patch_info[k]=patch_info['orig'].copy()
        include_patches[k]=np.stack([masks[k][x:x+patch_size,y:y+patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())]).mean((1,2))>=threshold
        patch_info[k]=patch_info[k][include_patches[k]]

        if no_break:
            patches[k]=np.stack([image[x:x+patch_size,y:y+patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())])
            np.save(os.path.join(dirname,"masks",f"{basename}_{k}.npy"),masks[k])
            np.save(os.path.join(dirname,"patches",f"{basename}_{k}.npy"),patches[include_patches])
            patch_info[k].to_pickle(os.path.join(dirname,"patches",f"{basename}_{k}.pkl"))
    if no_break: return None

    if not no_break:
        if df_section_pieces is not None: n_pieces=int(np.prod(df_section_pieces.loc[basename.replace("_ASAP","")]))
        G=radius_neighbors_graph(patch_info['tumor_map'][['x','y']], radius=512*np.sqrt(2))
        patch_info['tumor_map']['piece_ID']=connected_components(G)[1]
        if df_section_pieces is None: n_pieces=int(patch_info['tumor_map']['piece_ID'].max()+1)
        patch_info['tumor_map']['piece_ID']=patch_info['tumor_map']['piece_ID'].max()-patch_info['tumor_map']['piece_ID']
        patch_info['tumor_map']=patch_info['tumor_map'][patch_info['tumor_map']['piece_ID'].isin(patch_info['tumor_map']['piece_ID'].value_counts().index[:n_pieces].values)]
        patch_info['tumor_map']['piece_ID']=patch_info['tumor_map']['piece_ID'].map({v:k for k,v in enumerate(sorted(patch_info['tumor_map']['piece_ID'].unique()))})
        if df_section_pieces is not None:
            assert df_section_pieces.loc[basename.replace("_ASAP","")]['Pieces']<=2
            G=radius_neighbors_graph(patch_info['tumor_map'][['x','y']], radius=4096*np.sqrt(2))
            patch_info['tumor_map']['piece_ID']=patch_info['tumor_map']['piece_ID'].map(dict(zip(patch_info['tumor_map'].groupby("piece_ID")['x'].mean().sort_values(ascending=False).index,range(patch_info['tumor_map']['piece_ID'].max()+1))))
            patch_info['tumor_map']['section_ID']=connected_components(G)[1]
            complete=patch_info['tumor_map'][['section_ID','piece_ID']].groupby("section_ID")['piece_ID'].nunique()==df_section_pieces.loc[basename.replace("_ASAP","")]['Pieces']
            patch_info['tumor_map']['complete']=patch_info['tumor_map']['section_ID'].isin(complete[complete].index)
            while patch_info['tumor_map']['piece_ID'].max()+1<n_pieces:
                split_pieces=patch_info['tumor_map']['piece_ID'].value_counts().index
                for split_piece in split_pieces:
                    if patch_info['tumor_map'].loc[patch_info['tumor_map']['piece_ID']==split_piece,'complete'].sum()==0:
                        patch_info['tumor_map'].loc[patch_info['tumor_map']['piece_ID']==split_piece,'complete']=True # TODO: this can break
                        break
                G=radius_neighbors_graph(patch_info['tumor_map'].query(f"piece_ID=={split_piece}")[['x','y']], radius=patch_size*np.sqrt(2))
                cl=SpectralClustering(n_clusters=2,affinity="precomputed",assign_labels="discretize",eigen_solver="amg",n_components=2).fit_predict(G)
                patch_info['tumor_map'].loc[patch_info['tumor_map']['piece_ID']==split_piece,'piece_ID']=cl+patch_info['tumor_map']['piece_ID'].max()+1
                patch_info['tumor_map']['piece_ID']=patch_info['tumor_map']['piece_ID'].map(dict(zip(patch_info['tumor_map'].groupby("piece_ID")['x'].mean().sort_values(ascending=False).index,range(patch_info['tumor_map']['piece_ID'].max()+1))))
            patch_info['tumor_map']['section_ID']=patch_info['tumor_map']['piece_ID']//df_section_pieces.loc[basename.replace("_ASAP","")]['Pieces']
            assert patch_info['tumor_map']['piece_ID'].max()+1==n_pieces
        else:
            G=radius_neighbors_graph(patch_info['tumor_map'][['x','y']], radius=4096*np.sqrt(2))
            patch_info['tumor_map']['section_ID']=connected_components(G)[1]
            patch_info['tumor_map']['section_ID']=patch_info['tumor_map']['section_ID'].max()-patch_info['tumor_map']['section_ID']
        n_sections=patch_info['tumor_map']['section_ID'].max()+1
        n_pieces_per_section=(patch_info['tumor_map']['piece_ID'].max()+1)/n_sections
        pts=MultiPoint(patch_info['orig'][['x','y']].values)
        patch_info_new=[]
        for ID in patch_info['tumor_map']['piece_ID'].unique():
            tmp_points=patch_info['tumor_map'][['x','y']][patch_info['tumor_map']['piece_ID']==ID].values
            alpha_shape = alphashape.alphashape(tmp_points,alpha=alpha)
            tmp_points=MultiPoint(tmp_points)
            xy=dict()
            xy['macro']=pts.intersection(alpha_shape).difference(alpha_shape.exterior.buffer(256))
            xy['macro_tumor']=xy['macro'].intersection(tmp_points.buffer(64))
            xy['macro_no_tumor']=xy['macro'].difference(tmp_points.buffer(64))
            xy['tumor_no_macro']=tmp_points.difference(xy['macro'].buffer(64))
            for k in list(xy.keys()):
                if isinstance(xy[k],Point) or len(xy[k].geoms)==0:
                    del xy[k]
            del xy['macro']
            xy={k:pd.DataFrame(np.array([(int(p.x),int(p.y)) for p in xy[k]]),columns=['x','y']) for k in xy}
            for k in xy:
                xy[k]['basename']=basename
                xy[k]['section_ID']=ID//n_pieces_per_section
                xy[k]['piece_ID']=ID
                xy[k]['patch_size']=patch_size
                xy[k]['Type']=k
            xy=pd.concat(list(xy.values()),axis=0)
            xy['tumor_map']=xy['Type'].isin(['macro_tumor','tumor_no_macro'])
            xy['macro_map']=xy['Type'].isin(['macro_tumor','macro_no_tumor'])
            patch_info_new.append(xy)

        patch_info=pd.concat(patch_info_new,axis=0)
        for coord in ['x','y']: patch_info[f'{coord}_orig']=patch_info[coord]
        xy_bounds={}
        write_files=[]
        new_basenames=[]

        for ID in patch_info['section_ID' if use_section else 'piece_ID'].unique():
            new_basename=f"{basename}_{ID}"
            new_basenames.append(new_basename)
            include_patches=(patch_info['section_ID' if use_section else 'piece_ID']==ID).values
            patch_info_ID=patch_info[include_patches]
            (xmin,ymin),(xmax,ymax)=patch_info_ID[['x','y']].min(0).values,(patch_info_ID[['x','y']].max(0).values+patch_size)
            im=image[xmin:xmax,ymin:ymax]
            msk=masks['tumor_map'][xmin:xmax,ymin:ymax] # TODO: can break, need target tissue section
            patch_info_ID.loc[:,['x','y']]-=patch_info_ID[['x','y']].min(0)
            patches_ID=np.stack([im[x:x+patch_size,y:y+patch_size] for x,y in tqdm.tqdm(patch_info_ID[['x','y']].values.tolist())])
            patch_info_ID.reset_index(drop=True).to_pickle(os.path.join(dirname,"patches",f"{new_basename}.pkl"))
            write_files.append(dask.delayed(np.save)(os.path.join(dirname,"patches",f"{new_basename}.npy"),patches_ID))
            write_files.append(dask.delayed(np.save)(os.path.join(dirname,"masks",f"{new_basename}.npy"),cv2.resize(msk.astype(np.uint8),None,fx=1/image_mask_compression,fy=1/image_mask_compression,interpolation=cv2.INTER_NEAREST)>0 if image_mask_compression>1 else msk))
            if write_images:
                if image_mask_compression>1:
                    write_files.append(dask.delayed(np.save)(os.path.join(dirname,"images",f"{new_basename}.npy"),cv2.resize(im,None,fx=1/image_mask_compression,fy=1/image_mask_compression,interpolation=cv2.INTER_CUBIC)))
                else:
                    write_files.append(dask.delayed(np.save)(os.path.join(dirname,"images",f"{new_basename}.npy"),im))
            xy_bounds[ID]=((xmin,ymin),(xmax,ymax))
        pd.to_pickle(xy_bounds,os.path.join(dirname,"masks",f"{basename}.pkl"))
        with ProgressBar():
            dask.compute(write_files,scheduler='threading')
        pd.to_pickle(new_basenames,os.path.join(dirname,"metadata",f"{basename}.pkl"))
    return None

def preprocess_old(basename="163_A1a",
               threshold=0.05,
               patch_size=256,
               ext='.npy',
               secondary_patch_size=0):

    warnings.warn(
            "Old preprocessing is deprecated",
            DeprecationWarning
        )
    raise RuntimeError
    os.makedirs("masks",exist_ok=True)
    os.makedirs("patches",exist_ok=True)

    image=f"inputs/{basename}{ext}"
    basename=os.path.basename(image).replace(ext,'')
    image=load_image(image)#np.load(image)
    img_shape=image.shape[:-1]

    masks=dict()
    masks['tumor_map']=generate_tissue_mask(image,
                             compression=10,
                             otsu=False,
                             threshold=240,
                             connectivity=8,
                             kernel=5,
                             min_object_size=100000,
                             return_convex_hull=False,
                             keep_holes=False,
                             max_hole_size=6000,
                             gray_before_close=True,
                             blur_size=51)
    x_max,y_max=masks['tumor_map'].shape
    masks['macro_map']=fill_holes(masks['tumor_map'])

    patch_info=dict()
    for k in masks:
        patch_info[k]=pd.DataFrame([[basename,x,y,patch_size,"0"] for x,y in tqdm.tqdm(list(product(range(0,x_max-patch_size,patch_size),range(0,y_max-patch_size,patch_size))))],columns=['ID','x','y','patch_size','annotation'])
        patches=np.stack([image[x:x+patch_size,y:y+patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())])
        include_patches=np.stack([masks[k][x:x+patch_size,y:y+patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())]).mean((1,2))>=threshold

        np.save(f"masks/{basename}_{k}.npy",masks[k])
        np.save(f"patches/{basename}_{k}.npy",patches[include_patches])
        patch_info[k].iloc[include_patches].to_pickle(f"patches/{basename}_{k}.pkl")

    if secondary_patch_size:
        patch_info=dict()
        for k in ['tumor_map']:
            patch_info[k]=pd.DataFrame([[basename,x,y,secondary_patch_size,"0"] for x,y in tqdm.tqdm(list(product(range(0,x_max-secondary_patch_size,secondary_patch_size),range(0,y_max-secondary_patch_size,secondary_patch_size))))],columns=['ID','x','y','patch_size','annotation'])
            patches=np.stack([image[x:x+secondary_patch_size,y:y+secondary_patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())])
            include_patches=np.stack([masks[k][x:x+secondary_patch_size,y:y+secondary_patch_size] for x,y in tqdm.tqdm(patch_info[k][['x','y']].values.tolist())]).mean((1,2))>=threshold

            np.save(f"patches/{basename}_{k}_{secondary_patch_size}.npy",patches[include_patches])
            patch_info[k].iloc[include_patches].to_pickle(f"patches/{basename}_{k}_{secondary_patch_size}.pkl")
    return img_shape



"""
Ink Detection 
==========

Detect inks for tissue orientation.
"""
from skimage import morphology as morph
from scipy.ndimage import binary_opening, binary_dilation, label as scilabel
from skimage import filters, measure
from skimage.morphology import disk
import numpy as np, pandas as pd, copy
import sys,os,cv2
from itertools import product
from pathpretrain.utils import load_image
import warnings
import dask
from dask.diagnostics import ProgressBar
# sys.path.insert(0,os.path.abspath('.'))
from .filters import filter_red_pen, filter_blue_pen, filter_green_pen

def filter_yellow(img): # https://www.learnopencv.com/color-spaces-in-opencv-cpp-python/
    img_hsv=cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    return cv2.inRange(img_hsv,(10, 30, 30), (30, 255, 255))

ink_fn=dict(red=filter_red_pen,
           blue=filter_blue_pen,
           green=filter_green_pen,
           yellow=filter_yellow)

ink_min_size=dict(red=100,
           blue=30,
           green=30,
           yellow=1000)

colors=dict(red=np.array([255,0,0]),
           blue=np.array([0,0,255]),
           green=np.array([0,255,0]),
           yellow=np.array([255,255,0]))

def tune_mask(mask,edges,min_size=30):
    mask=(binary_dilation(mask,disk(3,bool),iterations=5) & edges)
    mask=binary_opening(mask,disk(3,bool),iterations=1)
    return morph.remove_small_objects(mask, min_size=min_size, connectivity = 2, in_place=True)>0

def filter_tune(img,color,edges):
    return tune_mask(~ink_fn[color](img),edges,min_size=ink_min_size[color])

def get_edges(mask):
    edges=filters.sobel(mask)>0
    edges = binary_dilation(edges,disk(30,bool))
    return edges

def detect_inks(basename="163_A1a",
                compression=8.,
                mask_compressed=True,
                ext=".npy",
                dirname=".",
                return_mean=False):
    """
    Detect inks in the specified image.

    Parameters
    ----------
    basename : str, optional
        The base name of the image to detect inks in, by default "163_A1a"
    compression : float, optional
        The compression factor to use when detecting inks, by default 8.
    mask_compressed : bool, optional
        Whether or not to use compressed masks, by default True
    ext : str, optional
        The file extension of the image, by default ".npy"
    dirname : str, optional
        The directory containing the image and masks, by default "."
    return_mean : bool, optional
        Whether or not to return the mean location of the detected inks, by default False

    Returns
    -------
    None
        The function saves the detected inks to a pickle file, but does not return anything.
    """

    os.makedirs(os.path.join(dirname,"detected_inks"),exist_ok=True)

    img=load_image(f"{dirname}/inputs/{basename}{ext}")
    xy_bounds=pd.read_pickle(os.path.join(dirname,"masks",f"{basename}.pkl"))
    with ProgressBar():
        masks=dask.compute({ID:dask.delayed(np.load)(f"{dirname}/masks/{basename}_{ID}.npy") for ID in xy_bounds},scheduler="threading")[0]
    pen_masks=dict()
    coord_translate={}
    for ID in xy_bounds:
        (xmin,ymin),(xmax,ymax)=xy_bounds[ID]
        msk=masks[ID].astype(np.uint8)
        if not mask_compressed: msk=cv2.resize(msk.astype(int),None,fx=1/compression,fy=1/compression,interpolation=cv2.INTER_NEAREST).astype(bool)
        coord_translate[ID]=np.array([xmin,ymin])
        im=cv2.resize(img[xmin:xmax,ymin:ymax],None,fx=1/compression,fy=1/compression)
        edges=dask.delayed(get_edges)(msk)#)
        com=np.vstack(np.where(msk)).T.mean(0)*compression
        pen_masks[ID]={color:dask.delayed(lambda x,y,z,k: np.vstack(np.where(z & filter_tune(x,k,y))).T*compression)(im,edges,msk,color) for color in ink_fn}
        pen_masks[ID]['center_mass']=com
    with ProgressBar():
        pen_masks=dask.compute(pen_masks,scheduler="threading")[0]
    pd.to_pickle(pen_masks,f"{dirname}/detected_inks/{basename}.pkl")
    return None


def detect_inks_old_v3(basename="163_A1a",
                compression=8,
                mask_compressed=True,
                ext=".npy",
                dirname=".",
                return_mean=False):

    os.makedirs(os.path.join(dirname,"detected_inks"),exist_ok=True)

    img=load_image(f"{dirname}/inputs/{basename}{ext}")
    xy_bounds=pd.read_pickle(os.path.join(dirname,"masks",f"{basename}.pkl"))
    img=cv2.resize(img,None,fx=1/compression,fy=1/compression)
    mask=np.zeros(img.shape[:-1],dtype=np.uint8)
    with ProgressBar():
        masks=dask.compute({ID:dask.delayed(np.load)(f"{dirname}/masks/{basename}_{ID}.npy") for ID in xy_bounds},scheduler="threading")[0]
    coord_translate={}
    for ID in xy_bounds:
        (xmin,ymin),(xmax,ymax)=xy_bounds[ID]
        coord_translate[ID+1]=np.array([xmin,ymin])
        msk=masks[ID].astype(np.uint8)#np.load(f"{dirname}/masks/{basename}_{ID}.npy").astype(np.uint8)
        if mask_compressed:
            xmin,ymin,xmax,ymax=(np.array([xmin,ymin,xmax,ymax])/compression).astype(int)
            xmax,ymax=np.array([xmin,ymin]+np.array(msk.shape))
        msk[msk>0]=ID+1
        mask[xmin:xmax,ymin:ymax]=msk
    if not mask_compressed: mask=cv2.resize(mask.astype(int),None,fx=1/compression,fy=1/compression,interpolation=cv2.INTER_NEAREST).astype(bool)
    labels,mask=mask,mask>0
    n_objects=np.max(labels.flatten())
    # labels,n_objects=scilabel(mask)
    edges=get_edges(mask)
    pen_masks={k:filter_tune(img,k,edges) for k in ink_fn}

    # for k in ['green','blue','red','yellow']:
    #     img[pen_masks[k],:]=colors[k]

    coords_df=pd.DataFrame(index=list(ink_fn.keys())+["center_mass"],columns=np.arange(1,n_objects+1))#
    for color,obj in product(coords_df.index[:-1],coords_df.columns):
        coords_df.loc[color,obj]=np.vstack(np.where((labels==obj) & (pen_masks[color]))).T*compression-coord_translate[obj]
        if return_mean: coords_df.loc[color,obj]=coords_df.loc[color,obj].mean(0)
    for obj in coords_df.columns:
        coords_df.loc["center_mass",obj]=np.vstack(np.where(labels==obj)).T.mean(0)*compression-coord_translate[obj]

    coords_df.to_pickle(f"{dirname}/detected_inks/{basename}.pkl")

def detect_inks_old_v2(basename="163_A1a_1",compression=8.,*args,**kwargs):

    os.makedirs("detected_inks",exist_ok=True)

    img=load_image(f"inputs/{basename}.npy")
    mask=np.load(f"masks/{basename}.npy")
    ID=int(basename.split("_")[-1])
    labels,mask=mask,labels>0
    edges=get_edges(mask)
    pen_masks={k:filter_tune(img,k,edges) for k in ink_fn}

    # for k in ['green','blue','red','yellow']:
    #     img[pen_masks[k],:]=colors[k]

    coords_df=pd.DataFrame(index=list(ink_fn.keys())+["center_mass"],columns=[ID])#
    coords_df.loc[color,ID]=np.vstack(np.where(mask & (pen_masks[color]))).T*compression
    coords_df.loc["center_mass",ID]=np.vstack(np.where(mask)).T.mean(0)*compression

    coords_df.to_pickle(f"detected_inks/{basename}.pkl")

def detect_inks_old(basename="163_A1a",
                compression=8,
                ext=".npy"):

    warnings.warn(
            "Old ink detection is deprecated",
            DeprecationWarning
        )
    raise RuntimeError

    os.makedirs("detected_inks",exist_ok=True)

    img,mask=load_image(f"inputs/{basename}{ext}"),np.load(f"masks/{basename}_macro_map.npy")
    img=cv2.resize(img,None,fx=1/compression,fy=1/compression)
    mask=cv2.resize(mask.astype(int),None,fx=1/compression,fy=1/compression,interpolation=cv2.INTER_NEAREST).astype(bool)
    labels,n_objects=scilabel(mask)
    edges=get_edges(mask)
    pen_masks={k:filter_tune(img,k,edges) for k in ink_fn}

    for k in ['green','blue','red','yellow']:
        img[pen_masks[k],:]=colors[k]

    coords_df=pd.DataFrame(index=list(ink_fn.keys())+["center_mass"],columns=np.arange(1,n_objects+1))
    for color,obj in product(coords_df.index[:-1],coords_df.columns):
        coords_df.loc[color,obj]=np.vstack(np.where((labels==obj) & (pen_masks[color]))).T*compression
    for obj in coords_df.columns:
        coords_df.loc["center_mass",obj]=np.vstack(np.where(labels==obj)).T.mean(0)*compression

    coords_df.to_pickle(f"detected_inks/{basename}.pkl")
    np.save(f"detected_inks/{basename}_thumbnail.npy",img)



"""
GNN 
==========

Graph neural network inference for tumor and completeness assessment.
"""
import os, torch, pickle, numpy as np, pandas as pd, torch.nn as nn
from torch_geometric.data import DataLoader as TG_DataLoader
from torch_geometric.utils import to_dense_batch, to_dense_adj, dense_to_sparse, dropout_adj, to_networkx
from torch_geometric.nn import GATConv
import torch.nn.functional as F

class GCNNet(torch.nn.Module):
    def __init__(self, inp_dim, out_dim, hidden_topology=[32,64,128,128], p=0.5, p2=0.1, drop_each=True):
        super(GCNNet, self).__init__()
        self.out_dim=out_dim
        self.convs = nn.ModuleList([GATConv(inp_dim, hidden_topology[0])]+[GATConv(hidden_topology[i],hidden_topology[i+1]) for i in range(len(hidden_topology[:-1]))])
        self.drop_edge = lambda edge_index: dropout_adj(edge_index,p=p2)[0]
        self.dropout = nn.Dropout(p)
        self.fc = nn.Linear(hidden_topology[-1], out_dim)
        self.drop_each=drop_each

    def forward(self, x, edge_index, edge_attr=None):
        for conv in self.convs:
            if self.drop_each and self.training: edge_index=self.drop_edge(edge_index)
            x = F.relu(conv(x, edge_index, edge_attr))
        if self.training:
            x = self.dropout(x)
        x = self.fc(x)
        return x

class GCNFeatures(torch.nn.Module):
    def __init__(self, gcn, bayes=False, p=0.05, p2=0.1):
        super(GCNFeatures, self).__init__()
        self.gcn=gcn
        self.drop_each=bayes
        self.gcn.drop_edge = lambda edge_index: dropout_adj(edge_index,p=p2)[0]
        self.gcn.dropout = nn.Dropout(p)

    def forward(self, x, edge_index, edge_attr=None):
        for i,conv in enumerate(self.gcn.convs):
            if self.drop_each: edge_index=self.gcn.drop_edge(edge_index)
            x = conv(x, edge_index, edge_attr)
            if i+1<len(self.gcn.convs):
                x=F.relu(x)
        if self.drop_each:
            x = self.gcn.dropout(x)
        y = self.gcn.fc(F.relu(x))#F.softmax()
        return x,y

def fix_state_dict(state_dict):
    # https://github.com/pyg-team/pytorch_geometric/issues/3139
    new_state_dict={}
    for k in state_dict:
        if '.att_' in k or '.lin_' in k:
            new_state_dict[k.replace("_l","_src").replace("_r","_dst")]=state_dict[k]
        else:
            new_state_dict[k]=state_dict[k]
    return new_state_dict

def predict(basename="163_A1a",
            analysis_type="tumor",
            gpu_id=0,
            dirname="."):
    """
    Run GNN prediction on patches.

    Parameters
    ----------
    basename : str
        Base name of the slide.
    analysis_type : str
        Type of analysis to run. Must be "tumor" or "macro".
    gpu_id : int, optional
        ID of the GPU to use. Default is 0.
    dirname : str, optional
        Directory to save results to. Default is current directory.

    Returns
    -------
    None
    """
    os.makedirs(os.path.join(dirname,"gnn_results"),exist_ok=True)
    hidden_topology=dict(tumor=[32,64,64],macro=[32,64,64])#[32]*3
    num_classes=dict(macro=4,tumor=3)
    if gpu_id>=0: torch.cuda.set_device(gpu_id)
    dataset=pickle.load(open(os.path.join(dirname,'graph_datasets',f"{basename}_{analysis_type}_map.pkl"),'rb'))
    model=GCNNet(dataset[0].x.shape[1],num_classes[analysis_type],hidden_topology=hidden_topology[analysis_type],p=0.,p2=0.)
    model.load_state_dict(fix_state_dict(torch.load(os.path.join(dirname,"models",f"{analysis_type}_map_gnn.pth"),map_location=f"cuda:{gpu_id}" if gpu_id>=0 else "cpu")))
    if torch.cuda.is_available():
        model=model.cuda()
    dataloader=TG_DataLoader(dataset,shuffle=False,batch_size=1)
    model.eval()
    feature_extractor=GCNFeatures(model,bayes=False)
    if torch.cuda.is_available():
        feature_extractor=feature_extractor.cuda()
    graphs=[]
    for i,data in enumerate(dataloader):
        with torch.no_grad():
            graph = to_networkx(data).to_undirected()
            model.train(False)
            x=data.x
            edge_index=data.edge_index
            if torch.cuda.is_available():
                x=x.cuda()
                edge_index=edge_index.cuda()
            xy=data.pos.numpy()
            preds=feature_extractor(x,edge_index)
            z,y_pred=preds[0].detach().cpu().numpy(),preds[1].detach().cpu().numpy()
            graphs.append(dict(G=graph,xy=xy,z=z,y_pred=y_pred,slide=data.id,component=data.component))
    torch.save(graphs,os.path.join(dirname,"gnn_results",f"{basename}_{analysis_type}_map.pkl"))



def tif2npy(in_file,out_dir,overwrite=False):
    import os, numpy as np
    import tifffile
    basename,ext=os.path.splitext(os.path.basename(in_file))
    out_file=os.path.join(out_dir,f"{basename}.npy")
    if not os.path.exists(out_file) or overwrite:
        np.save(out_file,tifffile.imread(in_file))

def display_results(out_graphs,res_,predict=False,custom_colors=[],s=1,img=None,alpha=None,scatter=True,scale=8,width_scale=20,node_scale=90,preds=None):
    import matplotlib
    import networkx as nx
    matplotlib.rcParams['figure.dpi']=300
    matplotlib.rcParams['axes.grid'] = False
    import matplotlib.pyplot as plt
    import cv2, numpy as np, pandas as pd
    # from arctic_ai.dgm.dgm.plotting import *
    import copy

    f = plt.figure(figsize=(15,15))
    ax = f.add_subplot(1, 1, 1)
    binary=False
    if not isinstance(img,type(None)): plt.imshow(np.transpose(img,(1,0,2)))

    for out_graph,res,pred in zip(out_graphs,res_,preds):
        xy=pred["xy"]
        y_orig=pred["y"]
        y=copy.deepcopy(y_orig)
        graph=out_graph
        node_color=res['mnode_to_color']; node_size=res['node_sizes']; edge_weight=res['edge_weight']
        if custom_colors: node_color=custom_colors
        node_list=res['node_list']; name='wsi'
        cmap = cm.coolwarm
        cmap = cm.get_cmap(cmap, 100)
        plt.set_cmap(cmap)
        edges = graph.edges()
        weights = np.array([edge_weight[(min(u, v), max(u, v))] for u, v in edges], dtype=np.float32)
        width = weights * width_scale
        node_size = np.sqrt(node_size) * node_scale
        c=y.flatten()
        pos = {}
        for node in graph.nodes():
            if len(res['mnode_to_nodes'][node])-1:
                pos[node]=np.array([xy[i] for i in res['mnode_to_nodes'][node]]).mean(0)/scale
            else:
                pos[node]=xy[list(res['mnode_to_nodes'][node])[0]]/scale
        if scatter: plt.scatter(xy[:,0]/scale,xy[:,1]/scale,c=c,alpha=alpha,s=s)
        nx.draw(graph, pos=pos, node_color=node_color, width=width, node_size=node_size,
                node_list=node_list, ax=ax, cmap=cmap)
    plt.axis('off')
    plt.grid(b=None)
    return None

def plot_results(basename="163_A1c",
                 compression=4,
                 k='macro'):
    import cv2, numpy as np, pandas as pd
    img=np.load(f"inputs/{basename}.npy")
    im=cv2.resize(img,None,fx=1/compression,fy=1/compression)
    mapper_graphs=pd.read_pickle(f"mapper_graphs/{basename}.pkl")
    for i in range(len(mapper_graphs[k])):
        mapper_graphs[k][i]['graph']['y']=mapper_graphs[k][i]['graph']['y_pred'].argmax(1)
    out_graphs,res_,preds=[mapper_graphs[k][i]['out_res'][0] for i in range(len(mapper_graphs[k]))],[mapper_graphs[k][i]['out_res'][1] for i in range(len(mapper_graphs[k]))],[mapper_graphs[k][i]['graph'] for i in range(len(mapper_graphs[k]))]
    display_results(out_graphs,res_,alpha=0.2,s=20,img=im,preds=preds,scale=compression,node_scale=30)

def try_except(fn,*args,**kwargs):
    try:
        fn(*args,**kwargs)
    except:
        pass

def return_osd_template():
    import os, subprocess
    try_except(os.makedirs,"dzi_files",exist_ok=True)
    for f in ['osd_scripts','dzi_files/openseadragon','dzi_files/style.css','./osd_template.html']:
        try_except(os.system,f"rm -rf ${f}")
    try_except(os.system,"git clone https://github.com/jlevy44/osd_scripts")
    try_except(os.link,"osd_scripts/osd_template.html","./osd_template.html")
    try_except(os.link,"osd_scripts/style.css","dzi_files/style.css")
    try_except(os.system,"cp -al osd_scripts/openseadragon dzi_files/")



"""
CLI 
==========

Contains functions for creating command line interfaces.
"""
import fire, os

class Commands(object):
    """A class for defining command line interface methods using Google Fire.

    Attributes:
        None
    """
    def __init__(self):
        pass

    def preprocess(self, 
                basename: str = "163_A1a", 
                threshold: float = 0.05, 
                patch_size: int = 256, 
                ext: str = ".npy", 
                secondary_patch_size: int = 0, 
                df_section_pieces_file: str = "", 
                image_mask_compression: float = 8., 
                dirname: str = ".") -> None:
        """
        Preprocesses an image and generates patches for training or testing.

        Args:
            basename: The base filename of the image to be preprocessed.
            threshold: The maximum fraction of a patch that can be blank. Default is 0.05.
            patch_size: The size of the patches to be generated. Default is 256.
            ext: The file extension of the input image. Default is ".npy".
            secondary_patch_size: The size of the patches for secondary processing. Default is 0.
            df_section_pieces_file: The filename of the file containing metadata about image patches. Default is "section_pieces.pkl".
            image_mask_compression: The degree of compression applied to the image mask. Default is 8.
            dirname: The directory where input and output files are stored. Default is ".".
            
        Returns:
            None
        """
        from arctic_ai.preprocessing import preprocess
        preprocess(basename,threshold,patch_size,ext,secondary_patch_size,df_section_pieces_file=df_section_pieces_file,image_mask_compression=image_mask_compression, dirname=dirname)

    def cnn_predict(self,
                basename: str = "163_A1a",
                analysis_type: str = "tumor",
                gpu_id: int = -1) -> None:
        """
        Generates embeddings from preprocessed images using a trained CNN model.
        
        Parameters
        ----------
        basename : str, optional
            The base filename of the preprocessed image, default is "163_A1a".
        analysis_type : str, optional
            The type of analysis to be performed, default is "tumor".
        gpu_id : int, optional
            The ID of the GPU to be used for prediction, default is -1 (CPU).
        """
        from arctic_ai.cnn_prediction import generate_embeddings
        generate_embeddings(basename,analysis_type,gpu_id)

    def graph_creation(self,
                   basename: str = "163_A1a",
                   analysis_type: str = "tumor",
                   radius: int = 256,
                   min_component_size: int = 600,
                   no_component_break: bool = True) -> None:
        """
        Generates a graph based on embeddings of image patches.

        Parameters
        ----------
        basename : str, optional
            The base filename of the image to be used for generating the graph. Default is "163_A1a".
        analysis_type : str, optional
            The type of analysis to be performed on the image. Default is "tumor".
        radius : int, optional
            The radius used for generating the graph. Default is 256.
        min_component_size : int, optional
            The minimum size for a connected component in the graph. Default is 600.
        no_component_break : bool, optional
            If True, the function will not break up large connected components. Default is True.

        Returns
        -------
        None
        """
        from arctic_ai.generate_graph import create_graph_data
        create_graph_data(basename,analysis_type,radius,min_component_size)

    def gnn_predict(self,
                      basename="163_A1a",
                      analysis_type="tumor",
                      radius=256,
                      min_component_size=600,
                      gpu_id=-1,
                      generate_graph=True,
                      no_component_break=True):
        """Run GNN prediction on patches.

        Parameters
        ----------
        basename : str, optional
            The base filename of the image to be segmented. Default is "163_A1a".
        analysis_type : str, optional
            The type of analysis to perform. Default is "tumor".
        radius : int, optional
            The radius used to generate the graphs. Default is 256.
        min_component_size : int, optional
            The minimum size of the connected components in graph. Default is 600.
        gpu_id : int, optional
            The ID of the GPU to use for prediction. Default is -1, which uses the CPU.
        generate_graph : bool, optional
            Whether to generate the graph data from the preprocessed image. Default is True.
        no_component_break : bool, optional
            Whether to avoid breaking connected components across patches. Default is True.

        Returns
        -------
        None
        """
        from arctic_ai.gnn_prediction import predict
        if generate_graph:
            from arctic_ai.generate_graph import create_graph_data
            create_graph_data(basename,analysis_type,radius,min_component_size,no_component_break=no_component_break)
        predict(basename,analysis_type,gpu_id)

    def nuclei_predict(self, 
                       predictor_dir="./", 
                       predictor_file="", 
                       patch_file="", 
                       threshold=0.05, 
                       savenpy='pred.npy') -> None:
        '''
        The nuclei_predict function takes in a set of arguments including predictor_dir, predictor_file, patch_file, threshold, and savenpy. It loads a model from the specified predictor_dir and predictor_file, runs it on a set of patches in patch_file, and saves the output predictions as an npy stack and label dictionary if savenpy is not None. If savexml is not None, it also saves predictions to an ASAP xml.
        The function loads the model, flips the patches, runs the model on the patches, and saves the predictions in the specified formats if requested.
        Args:
            predictor_dir (str): path to model folder
            predictor_file (str): filename of model in the folder (don't include path to folder)
            patch_file (str): path to an npy stack of patches
            classifier_type (BasePredictor): class (from predict.py)
            panoptic (bool): whether the model performs panoptic segmentation. If false, it is assumed to do instance segmentation.
            n (int): number of classes the model classifies into 
            threshold (float): threshold to use for model 
            savenpy (str): Path to file to which to save npy output. The output is a numpy stack of the the masks for each patch. In the mask, nuclei are given a non-zero integer, the ID of the the instance it is a part of. A pickled dictionary mapping the instance ID to the class label is also outputted. If savenpy=None, predictions are not saved in an npy format. 
            savexml (str): Path to file to which to save xml output (ASAP format). If savexml=None, predictions are not saved in an xml format. 
            patch_coords (str): an extra pkl file which specifies the x,y (x is row, y is col) metadata for the patches. Must be provided if exporting to xml, since location is a part of the ASAP format
        '''
        # raise NotImplementedError("Removed temporarily, adding back soon")
        from arctic_ai.detection_workflows.nuclei_cli import detect_from_patches
        detect_from_patches(predictor_dir=predictor_dir, 
                       predictor_file=predictor_file, 
                       patch_file=patch_file, 
                       threshold=threshold, 
                       savenpy=savenpy)

    def quality_score(self,
                      basename="163_A1a"):
        from arctic_ai.quality_scores import generate_quality_scores
        generate_quality_scores(basename)

    def ink_detect(self,
                   basename="163_A1a",
                   compression=8):
        """
        Detect inks in the specified image.

        Parameters
        ----------
        basename : str, optional
            The base name of the image to detect inks in, by default "163_A1a"
        compression : float, optional
            The compression factor to use when detecting inks, by default 8.

        Returns
        -------
        None
            The function saves the detected inks to a pickle file, but does not return anything.
        """
        from arctic_ai.ink_detection import detect_inks
        detect_inks(basename,compression)

    def dump_results(self,
                     patient="163_A1",
                     scheme="2/1"):
        raise NotImplementedError("Deprecated")
        from arctic_ai.compile_results import dump_results
        dump_results(patient,scheme)

    def run_series_old(self,
                   patient="163_A1",
                   input_dir="inputs",
                   scheme="2/1",
                   compression=1.,
                   overwrite=True,
                   record_time=False,
                   extract_dzi=False,
                   ext=".npy"):
        raise NotImplementedError("Deprecated")
        from arctic_ai.workflow import run_series
        run_series(patient,input_dir,scheme,compression,overwrite,record_time,extract_dzi,ext)

    def run_series(self,
                   patient="163_A1",
                   input_dir="inputs",
                   compression=1.,
                   overwrite=True,
                   record_time=False,
                   ext=".npy",
                   dirname=".",
                   df_section_pieces_file="df_section_pieces.pkl",
                   run_stitch_slide=True):
        """Runs the entire image analysis workflow on a given patient.

        Parameters
        ----------
        patient : str, optional
            The patient ID to be processed. Default is "163_A1".
        input_dir : str, optional
            The directory where the input files are stored. Default is "inputs".
        compression : float, optional
            The degree of compression applied to the input image. Default is 1.0.
        overwrite : bool, optional
            If True, overwrite existing output files. Default is True.
        record_time : bool, optional
            If True, record the time taken for each step of the workflow. Default is False.
        ext : str, optional
            The file extension of the input image. Default is ".npy".
        dirname : str, optional
            The directory where input and output files are stored. Default is ".".
        df_section_pieces_file : str, optional
            The filename of the file containing metadata about image patches. Default is "df_section_pieces.pkl".
        This method runs the entire image analysis workflow on a given patient, with options to customize the input and output directories, degree of image compression, file extension, and whether to overwrite existing output files. If record_time is set to True, the time taken for each step of the workflow will be recorded. The default values for patient, input_dir, compression, overwrite, ext, dirname, and df_section_pieces_file are "163_A1", "inputs", 1.0, True, ".npy", ".", and "df_section_pieces.pkl", respectively.
        """
        from arctic_ai.workflow import run_series
        run_series(patient,input_dir,compression,overwrite,record_time,ext,dirname,df_section_pieces_file,run_stitch_slide)

    def run_parallel(self,
                    patient="163_A1",
                    input_dir="inputs",
                    compression=1.,
                    overwrite=True,
                    record_time=False,
                    ext=".npy",
                    dirname=".",
                    df_section_pieces_file="df_section_pieces.pkl",
                    run_stitch_slide=True):
            """Runs the entire image analysis workflow on a given patient, in parallel using either a local executor or slurm (more executors will be added).

            Parameters
            ----------
            patient : str, optional
                The patient ID to be processed. Default is "163_A1".
            input_dir : str, optional
                The directory where the input files are stored. Default is "inputs".
            compression : float, optional
                The degree of compression applied to the input image. Default is 1.0.
            overwrite : bool, optional
                If True, overwrite existing output files. Default is True.
            record_time : bool, optional
                If True, record the time taken for each step of the workflow. Default is False.
            ext : str, optional
                The file extension of the input image. Default is ".npy".
            dirname : str, optional
                The directory where input and output files are stored. Default is ".".
            df_section_pieces_file : str, optional
                The filename of the file containing metadata about image patches. Default is "df_section_pieces.pkl".
            This method runs the entire image analysis workflow on a given patient, with options to customize the input and output directories, degree of image compression, file extension, and whether to overwrite existing output files. If record_time is set to True, the time taken for each step of the workflow will be recorded. The default values for patient, input_dir, compression, overwrite, ext, dirname, and df_section_pieces_file are "163_A1", "inputs", 1.0, True, ".npy", ".", and "df_section_pieces.pkl", respectively.
            """
            from arctic_ai.scale_workflow import run_parallel
            run_parallel(patient,input_dir,compression,overwrite,record_time,ext,dirname,df_section_pieces_file,run_stitch_slide)

    def tif2npy(self,
                in_file='',
                out_dir='./'):
        """Converts a .tif file to a .npy file.

        Parameters
        ----------
        in_file : str
            The path to the input .tif file.
        out_dir : str
            The path to the output directory where the .npy file will be saved.
        """
        from arctic_ai.utils import tif2npy
        tif2npy(in_file,out_dir)

    def dzi_folder_setup(self):
        from arctic_ai.utils import return_osd_template
        return_osd_template()

    def im2dzi(self,
               in_file='',
               out_dir='./',
               compression: float = 1.):
        """
        Converts an input image file to Deep Zoom Image format.

        Parameters
        ----------
        in_file : str
            The path to the input image file.
        out_dir : str
            The output directory where the DZI file will be written.
            Default is "./".
        compression : float
            The degree of compression applied to the DZI file. Default is 1.

        Returns
        -------
        None
        """
        from arctic_ai.image_stitch import npy2dzi
        basename,_=os.path.splitext(os.path.basename(in_file))
        npy2dzi(npy_file=in_file,
                    dzi_out=os.path.join(out_dir,f"{basename}.dzi"),
                    compression=compression)

    def write_dzis(self,
                   basename="",
                    compression=4,
                    dirname=".",
                    ext=".tif"):
        raise NotImplementedError("Deprecated")
        from arctic_ai.image_stitch import stitch_slides
        stitch_slides(basename=basename,
                     compression=compression,
                     dirname=dirname,
                     ext=ext)


    def extract_dzis(self,
                     patient='163_A1',
                     overwrite_scheme='',
                     types=['image','tumor','macro']):
        raise NotImplementedError("Deprecated")
        from arctic_ai.case_prototype import Case
        case=Case(patient=patient,overwrite_scheme=overwrite_scheme)
        for k in types:
            case.extract2dzi(k)

def main():
    fire.Fire(Commands)

if __name__=="__main__":
    main()



from setuptools import setup
from setuptools.command.install import install
import subprocess
import os

with open("requirements.txt") as f:
      PACKAGES=f.read().splitlines()

with open('README.md','r', encoding='utf-8') as f:
      long_description = f.read()

setup(name='arctic_ai',
      version='0.1',
      description='Python package for Arctic Workflow. Mirrors jupyter developments.',
      url='https://github.com/jlevy44/ArcticAI_Prototype',
      author='Joshua Levy',
      author_email='joshualevy44@berkeley.edu',
      license='MIT',
      scripts=[],
      entry_points={
            'console_scripts':['arctic_ai=arctic_ai.cli:main']
      },
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=['arctic_ai'],
      install_requires=PACKAGES)



"""
CNN 
==========

Contains functions related to generating embeddings for image patches using a convolutional neural network
"""
import os, torch, tqdm, pandas as pd, numpy as np
from torch.utils.data import Dataset, DataLoader
from PIL import Image
from pathpretrain.train_model import train_model, generate_transformers, generate_kornia_transforms
import warnings

class CustomDataset(Dataset):
    # load using saved patches and mask file
    def __init__(self, ID, patch_info, X, transform):
        self.X=X
        self.patch_info=patch_info
        self.xy=self.patch_info[['x','y']].values
        self.patch_size=self.patch_info['patch_size'].iloc[0]
        self.length=self.patch_info.shape[0]
        self.transform=transform
        self.to_pil=lambda x: Image.fromarray(x)
        self.ID=ID#os.path.basename(npy_file).replace(".npy","")

    def __getitem__(self,i):
        x,y=self.xy[i]
        return self.transform(self.to_pil(self.X[i]))#[x:x+patch_size,y:y+patch_size]

    def __len__(self):
        return self.length

    def embed(self,model,batch_size,out_dir):
        Z=[]
        dataloader=DataLoader(self,batch_size=batch_size,shuffle=False)
        n_batches=len(self)//batch_size
        with torch.no_grad():
            for i,X in tqdm.tqdm(enumerate(dataloader),total=n_batches):
                if torch.cuda.is_available(): X=X.cuda()
                z=model(X).detach().cpu().numpy()
                Z.append(z)
        Z=np.vstack(Z)
        torch.save(dict(embeddings=Z,patch_info=self.patch_info),os.path.join(out_dir,f"{self.ID}.pkl"))

class CustomDatasetOld(Dataset):
    # load using saved patches and mask file
    def __init__(self, patch_info, npy_file, transform):
        warnings.warn(
                "This dataset class is deprecated.",
                DeprecationWarning
            )
        raise RuntimeError
        self.X=np.load(npy_file)
        self.patch_info=pd.read_pickle(patch_info)
        self.xy=self.patch_info[['x','y']].values
        self.patch_size=self.patch_info['patch_size'].iloc[0]
        self.length=self.patch_info.shape[0]
        self.transform=transform
        self.to_pil=lambda x: Image.fromarray(x)
        self.ID=os.path.basename(npy_file).replace(".npy","")

    def __getitem__(self,i):
        x,y=self.xy[i]
        return self.transform(self.to_pil(self.X[i]))#[x:x+patch_size,y:y+patch_size]

    def __len__(self):
        return self.length

    def embed(self,model,batch_size,out_dir):
        Z=[]
        dataloader=DataLoader(self,batch_size=batch_size,shuffle=False)
        n_batches=len(self)//batch_size
        with torch.no_grad():
            for i,X in tqdm.tqdm(enumerate(dataloader),total=n_batches):
                if torch.cuda.is_available(): X=X.cuda()
                z=model(X).detach().cpu().numpy()
                Z.append(z)
        Z=np.vstack(Z)
        torch.save(dict(embeddings=Z,patch_info=self.patch_info),os.path.join(out_dir,f"{self.ID}.pkl"))

def generate_embeddings(basename="163_A1a",
                        analysis_type="tumor",
                       gpu_id=0,
                       dirname="."):
    """ 
    Generate embeddings for patches in a WSI.

    Parameters
    ----------
    basename : str
        Basename of the WSI.
    analysis_type : str
        Type of analysis to perform. Can be either "tumor" or "macro".
    gpu_id : int, optional
        GPU to use for training. If not provided, uses CPU.
    dirname : str, optional
        Directory containing data for the WSI.

    Returns
    -------
    None
        The function saves the generated embeddings to the `cnn_embeddings` directory.
    """

    os.makedirs(os.path.join(dirname,"cnn_embeddings"),exist_ok=True)

    patch_info_file,npy_file=os.path.join(dirname,f"patches/{basename}.pkl"),os.path.join(dirname,f"patches/{basename}.npy")
    models={k:os.path.join(dirname,f"models/{k}_map_cnn.pth") for k in ['macro','tumor']}
    num_classes=dict(macro=4,tumor=3)

    npy_stack=np.load(npy_file)
    patch_info=pd.read_pickle(patch_info_file)
    if f"{analysis_type}_map" in patch_info.columns:
        npy_stack=npy_stack[patch_info[f"{analysis_type}_map"].values]
        patch_info=patch_info[patch_info[f"{analysis_type}_map"].values]
    train_model(model_save_loc=models[analysis_type],extract_embeddings=True,num_classes=num_classes[analysis_type],predict=True,embedding_out_dir=os.path.join(dirname,"cnn_embeddings/"),custom_dataset=CustomDataset(f"{basename}_{analysis_type}_map",patch_info,npy_stack,generate_transformers(224,256)['test']),gpu_id=gpu_id)


def generate_embeddings_old(basename="163_A1a",
                        analysis_type="tumor",
                       gpu_id=0,
                       dirname="."):

    warnings.warn(
            "Old generate embeddings function is deprecated",
            DeprecationWarning
        )
    raise RuntimeError

    os.makedirs(os.path.join(dirname,"cnn_embeddings"),exist_ok=True)

    patch_info_file,npy_file=os.path.join(dirname,f"patches/{basename}_{analysis_type}_map.pkl"),os.path.join(dirname,f"patches/{basename}_{analysis_type}_map.npy")
    models={k:f"models/{k}_map_cnn.pth" for k in ['macro','tumor']}
    num_classes=dict(macro=4,tumor=3)
    train_model(model_save_loc=models[analysis_type],extract_embeddings=True,num_classes=num_classes[analysis_type],predict=True,embedding_out_dir=os.path.join(dirname,"cnn_embeddings/"),custom_dataset=CustomDataset(patch_info_file,npy_file,generate_transformers(224,256)['test']),gpu_id=gpu_id)
